#!/usr/bin/env python3

import argparse
import binascii
import ctypes
import enum
from collections import defaultdict

from typing_extensions import Self


class ValidationError(Exception):
    """Generic patch validation exception"""


class OpCode(enum.IntEnum):
    """Patch file operation code"""

    COPY_LEN_U4 = 0 << 4
    COPY_LEN_U12 = 1 << 4
    COPY_LEN_U20 = 2 << 4
    COPY_LEN_U32 = 3 << 4
    WRITE_LEN_U4 = 4 << 4
    WRITE_LEN_U12 = 5 << 4
    WRITE_LEN_U20 = 6 << 4
    WRITE_LEN_U32 = 7 << 4
    ADDR_SHIFT_S8 = 8 << 4
    ADDR_SHIFT_S16 = 9 << 4
    ADDR_SET_U32 = 10 << 4
    PATCH = 11 << 4
    OPCODE_MASK = 0xF0
    DATA_MASK = 0x0F

    @classmethod
    def from_byte(cls, byte: int):
        return cls(byte & cls.OPCODE_MASK)

    @classmethod
    def data(cls, byte: int):
        return byte & cls.DATA_MASK


class Instr:
    """Parent instruction class"""

    def ctypes_class(self):
        """Instruction ctypes class"""
        raise NotImplementedError

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    def __len__(self):
        return ctypes.sizeof(self.ctypes_class())

    @classmethod
    def from_bytes(
        cls,
        b: bytes,
        offset: int,
        original_offset: int,
    ):
        """Reconstruct class from bytes"""
        opcode = OpCode.from_byte(b[offset])
        if (
            opcode == OpCode.COPY_LEN_U4
            or opcode == OpCode.COPY_LEN_U12
            or opcode == OpCode.COPY_LEN_U20
            or opcode == OpCode.COPY_LEN_U32
        ):
            return CopyInstr.from_bytes(b, offset, original_offset)
        if (
            opcode == OpCode.WRITE_LEN_U4
            or opcode == OpCode.WRITE_LEN_U12
            or opcode == OpCode.WRITE_LEN_U20
            or opcode == OpCode.WRITE_LEN_U32
        ):
            return WriteInstr.from_bytes(b, offset, original_offset)
        if opcode == OpCode.ADDR_SHIFT_S8 or opcode == OpCode.ADDR_SHIFT_S16 or opcode == OpCode.ADDR_SET_U32:
            return SetAddrInstr.from_bytes(b, offset, original_offset)
        if opcode == OpCode.PATCH:
            return PatchInstr.from_bytes(b, offset, original_offset)

        raise NotImplementedError


class SetAddrInstr(Instr):
    class ShiftAddrS8(ctypes.LittleEndianStructure):
        op = OpCode.ADDR_SHIFT_S8
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("val", ctypes.c_int8),
        ]
        _pack_ = 1

    class ShiftAddrS16(ctypes.LittleEndianStructure):
        op = OpCode.ADDR_SHIFT_S16
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("val", ctypes.c_int16),
        ]
        _pack_ = 1

    class SetAddrU32(ctypes.LittleEndianStructure):
        op = OpCode.ADDR_SET_U32
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("val", ctypes.c_uint32),
        ]
        _pack_ = 1

    def __init__(self, old_addr, new_addr, cls_override=None):
        self.old = old_addr
        self.new = new_addr
        self.shift = self.new - self.old
        self._cls_override = cls_override

    def ctypes_class(self):
        if self._cls_override is not None:
            return self._cls_override

        if -128 <= self.shift <= 127:
            return self.ShiftAddrS8
        elif -32768 <= self.shift <= 32767:
            return self.ShiftAddrS16
        else:
            return self.SetAddrU32

    @classmethod
    def from_bytes(cls, b: bytes, offset: int, original_offset: int) -> tuple[Self, int, int]:
        opcode = b[offset]
        if opcode == OpCode.ADDR_SHIFT_S8:
            s8 = cls.ShiftAddrS8.from_buffer_copy(b, offset)
            c = cls(original_offset, original_offset + s8.val)
            struct_len = ctypes.sizeof(s8)
        elif opcode == OpCode.ADDR_SHIFT_S16:
            s16 = cls.ShiftAddrS16.from_buffer_copy(b, offset)
            c = cls(original_offset, original_offset + s16.val)
            struct_len = ctypes.sizeof(s16)
        elif opcode == OpCode.ADDR_SET_U32:
            s32 = cls.SetAddrU32.from_buffer_copy(b, offset)
            c = cls(original_offset, s32.val)
            struct_len = ctypes.sizeof(s32)
        else:
            raise RuntimeError
        return c, struct_len, c.new

    def __bytes__(self):
        instr = self.ctypes_class()
        if instr == self.ShiftAddrS8 or instr == self.ShiftAddrS16:
            val = self.shift
        else:
            val = self.new

        return bytes(instr(instr.op.value, val))

    def __str__(self):
        if -32768 <= self.shift <= 32767:
            return f" ADDR: shifting {self.shift} (from {self.old:08x} to {self.new:08x})"
        else:
            return f" ADDR: now {self.new:08x} (shift of {self.new - self.old})"


class CopyInstr(Instr):
    class CopyGeneric(ctypes.LittleEndianStructure):
        pass

    class CopyU4(CopyGeneric):
        op = OpCode.COPY_LEN_U4
        _fields_ = [
            ("opcode", ctypes.c_uint8),
        ]
        _pack_ = 1

        @property
        def length(self):
            return OpCode.data(self.opcode)

    class CopyU12(CopyGeneric):
        op = OpCode.COPY_LEN_U12
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("_length", ctypes.c_uint8),
        ]
        _pack_ = 1

        @property
        def length(self):
            return (OpCode.data(self.opcode) << 8) | self._length

    class CopyU20(CopyGeneric):
        op = OpCode.COPY_LEN_U20
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("_length", ctypes.c_uint16),
        ]
        _pack_ = 1

        @property
        def length(self):
            return (OpCode.data(self.opcode) << 16) | self._length

    class CopyU32(CopyGeneric):
        op = OpCode.COPY_LEN_U32
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("length", ctypes.c_uint32),
        ]
        _pack_ = 1

    def __init__(self, length: int, original_offset: int = -1, cls_override=None):
        assert length != 0
        self.length = length
        # Used in construction to simplify optimisations
        self.original_offset = original_offset
        self._cls_override = cls_override

    def ctypes_class(self):
        if self._cls_override is not None:
            return self._cls_override
        if self.length < 16:
            return self.CopyU4
        elif self.length < 4096:
            return self.CopyU12
        elif self.length < 1048576:
            return self.CopyU20
        else:
            return self.CopyU32

    @classmethod
    def _from_opcode(cls, op: OpCode) -> type[CopyGeneric]:
        if op == OpCode.COPY_LEN_U4:
            return cls.CopyU4
        elif op == OpCode.COPY_LEN_U12:
            return cls.CopyU12
        elif op == OpCode.COPY_LEN_U20:
            return cls.CopyU20
        elif op == OpCode.COPY_LEN_U32:
            return cls.CopyU32
        else:
            raise RuntimeError

    @classmethod
    def from_bytes(cls, b: bytes, offset: int, original_offset: int) -> tuple[Self, int, int]:
        opcode = OpCode.from_byte(b[offset])
        op_class = cls._from_opcode(opcode)
        s = op_class.from_buffer_copy(b, offset)
        return cls(s.length), ctypes.sizeof(s), original_offset + s.length

    def __bytes__(self):
        instr = self.ctypes_class()
        if instr == self.CopyU4:
            return bytes(instr(instr.op.value | self.length))
        elif instr == self.CopyU12:
            top = self.length >> 8
            bottom = self.length & 0xFF
            return bytes(instr(instr.op.value | top, bottom))
        elif instr == self.CopyU20:
            top = self.length >> 16
            bottom = self.length & 0xFFFF
            return bytes(instr(instr.op.value | top, bottom))
        else:
            return bytes(instr(instr.op.value, self.length))

    def __str__(self):
        return f" COPY: {self.length:6d} bytes"


class WriteInstr(Instr):
    class WriteGeneric(ctypes.LittleEndianStructure):
        pass

    class WriteU4(WriteGeneric):
        op = OpCode.WRITE_LEN_U4
        _fields_ = [
            ("opcode", ctypes.c_uint8),
        ]
        _pack_ = 1

        @property
        def length(self):
            return OpCode.data(self.opcode)

    class WriteU12(WriteGeneric):
        op = OpCode.WRITE_LEN_U12
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("_length", ctypes.c_uint8),
        ]
        _pack_ = 1

        @property
        def length(self):
            return (OpCode.data(self.opcode) << 8) | self._length

    class WriteU20(WriteGeneric):
        op = OpCode.WRITE_LEN_U20
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("_length", ctypes.c_uint16),
        ]
        _pack_ = 1

        @property
        def length(self):
            return (OpCode.data(self.opcode) << 16) | self._length

    class WriteU32(WriteGeneric):
        op = OpCode.WRITE_LEN_U32
        _fields_ = [
            ("opcode", ctypes.c_uint8),
            ("length", ctypes.c_uint32),
        ]
        _pack_ = 1

    def __init__(self, data, cls_override=None):
        assert len(data) != 0
        self.data = data
        self._cls_override = cls_override

    def ctypes_class(self):
        if self._cls_override is not None:
            return self._cls_override
        if len(self.data) < 16:
            return self.WriteU4
        elif len(self.data) < 4096:
            return self.WriteU12
        elif len(self.data) < 1048576:
            return self.WriteU20
        else:
            return self.WriteU32

    @classmethod
    def _from_opcode(cls, op: OpCode) -> type[WriteGeneric]:
        if op == OpCode.WRITE_LEN_U4:
            return cls.WriteU4
        elif op == OpCode.WRITE_LEN_U12:
            return cls.WriteU12
        elif op == OpCode.WRITE_LEN_U20:
            return cls.WriteU20
        elif op == OpCode.WRITE_LEN_U32:
            return cls.WriteU32
        else:
            raise RuntimeError

    @classmethod
    def from_bytes(cls, b: bytes, offset: int, original_offset: int) -> tuple[Self, int, int]:
        opcode = OpCode.from_byte(b[offset])
        op_class = cls._from_opcode(opcode)
        s = op_class.from_buffer_copy(b, offset)
        hdr_len = ctypes.sizeof(s)

        return (
            cls(b[offset + hdr_len : offset + hdr_len + s.length]),
            hdr_len + s.length,
            original_offset + s.length,
        )

    def __bytes__(self):
        instr = self.ctypes_class()
        if instr == self.WriteU4:
            return bytes(instr(instr.op.value | len(self.data))) + self.data
        elif instr == self.WriteU12:
            top = len(self.data) >> 8
            bottom = len(self.data) & 0xFF
            return bytes(instr(instr.op.value | top, bottom)) + self.data
        elif instr == self.WriteU20:
            top = len(self.data) >> 16
            bottom = len(self.data) & 0xFFFF
            return bytes(instr(instr.op.value | top, bottom)) + self.data
        else:
            return bytes(instr(instr.op.value, len(self.data))) + self.data

    def __str__(self):
        if len(self.data) < 64:
            return f"WRITE: {len(self.data):6d} bytes ({self.data.hex()})"
        else:
            return f"WRITE: {len(self.data):6d} bytes ({self.data[:64].hex()}...)"

    def __len__(self):
        return ctypes.sizeof(self.ctypes_class()) + len(self.data)


class PatchInstr(Instr):
    class PatchData(ctypes.LittleEndianStructure):
        op = OpCode.PATCH
        _fields_ = [("opcode", ctypes.c_uint8)]
        _pack_ = 1

    def __init__(self, operations):
        self.operations = operations

    def ctypes_class(self):
        return self.PatchData

    @classmethod
    def from_bytes(cls, b: bytes, offset: int, original_offset: int):
        assert b[offset] == OpCode.PATCH
        operations: list[Instr] = []
        length = 1

        while True:
            copy_len = b[offset + length]
            length += 1
            if copy_len == 0:
                break
            assert copy_len != 0
            actual_copy_len = copy_len & 0x7F
            original_offset += actual_copy_len
            operations.append(CopyInstr(int(actual_copy_len)))

            if copy_len & 0x80:
                write_len = 1
            else:
                write_len = b[offset + length]
                length += 1
            if write_len == 0:
                break
            assert write_len != 0
            original_offset += write_len
            operations.append(WriteInstr(b[offset + length : offset + length + write_len]))
            length += write_len

        return cls(operations), length, original_offset

    def __bytes__(self):
        x = OpCode.PATCH.value.to_bytes(1, "little")
        op_iterator = (o for o in self.operations)

        while True:
            copy_op = next(op_iterator, None)
            write_op = next(op_iterator, None)

            if copy_op is None:
                break
            assert isinstance(copy_op, CopyInstr)
            assert copy_op.length < 128
            assert copy_op.length != 0

            if write_op is not None and len(write_op.data) == 1:
                val = 0x80 | copy_op.length
            else:
                val = copy_op.length
            x += val.to_bytes(1, "little")

            if write_op is None:
                break
            assert isinstance(write_op, WriteInstr)
            assert len(write_op.data) < 256
            assert len(write_op.data) != 0
            if len(write_op.data) > 1:
                x += len(write_op.data).to_bytes(1, "little")
            x += write_op.data
        x += b"\x00"
        return x

    def __str__(self):
        return "PATCH:\n" + "\n".join([f"\t{str(o)}" for o in self.operations])

    def __len__(self):
        length = 2
        for op in self.operations:
            if isinstance(op, CopyInstr):
                length += 1
            elif isinstance(op, WriteInstr):
                if len(op.data) > 1:
                    length += 1
                length += len(op.data)
        return length


class cpatch:
    class PatchHeader(ctypes.LittleEndianStructure):
        VERSION_MAJOR = 1
        VERSION_MINOR = 0

        class ArrayValidation(ctypes.LittleEndianStructure):
            _fields_ = [
                ("length", ctypes.c_uint32),
                ("crc", ctypes.c_uint32),
            ]
            _pack_ = 1

        magic_value = 0xBA854092
        cache_size = 128
        _fields_ = [
            ("magic", ctypes.c_uint32),
            ("version_major", ctypes.c_uint8),
            ("version_minor", ctypes.c_uint8),
            ("original_file", ArrayValidation),
            ("constructed_file", ArrayValidation),
            ("patch_file", ArrayValidation),
            ("header_crc", ctypes.c_uint32),
        ]
        _pack_ = 1

    @classmethod
    def _naive_diff(cls, old: bytes, new: bytes, hash_len: int = 8):
        """Construct basic runs of WRITE, COPY, and SET_ADDR instructions"""
        instr: list[Instr] = []
        old_offset = 0
        new_offset = 0
        write_start = 0
        write_pending = 0

        # Pre-hash original image
        pre_hash = {}
        prev_val = None
        for offset in range(len(old) - hash_len):
            val = old[offset : offset + hash_len]
            if val == prev_val:
                continue
            if val not in pre_hash:
                pre_hash[val] = [offset]
            else:
                pre_hash[val].append(offset)
            prev_val = val

        # Loop until entire image is reconstructed
        while new_offset < len(new):
            val = new[new_offset : new_offset + hash_len]

            # If word exists in original image
            if val in pre_hash:
                if write_pending:
                    instr.append(WriteInstr(new[write_start : write_start + write_pending]))
                    write_pending = 0

                old_match = -100

                # Check to see if we have a match at current pointer
                if old_offset in pre_hash[val]:
                    old_match = 0
                    while (
                        (new_offset + old_match) < len(new)
                        and (old_offset + old_match) < len(old)
                        and new[new_offset + old_match] == old[old_offset + old_match]
                    ):
                        old_match += 1

                max_match = old_match
                max_offset = old_offset

                # For each location in original image
                for orig_offset in pre_hash[val]:
                    this_match = 0
                    while (
                        (new_offset + this_match) < len(new)
                        and (orig_offset + this_match) < len(old)
                        and new[new_offset + this_match] == old[orig_offset + this_match]
                    ):
                        this_match += 1

                    if this_match > max_match and this_match > (old_match + 8):
                        max_match = this_match
                        max_offset = orig_offset

                if max_offset != old_offset:
                    # Update memory address
                    instr.append(SetAddrInstr(old_offset, max_offset))

                instr.append(CopyInstr(max_match, max_offset))
                new_offset += max_match
                old_offset = max_offset + max_match
            else:
                if write_pending == 0:
                    write_start = new_offset
                write_pending += 1
                new_offset += 1
                old_offset += 1

        if write_pending:
            instr.append(WriteInstr(new[write_start : write_start + write_pending]))
            write_pending = 0

        return instr

    @classmethod
    def _cleanup_jumps(cls, old: bytes, instructions: list[Instr]) -> list[Instr]:
        """Find locations that jumped backwards just to jump forward to original location"""

        merged: list[Instr] = []
        while len(instructions) > 0:
            instr = instructions.pop(0)
            replaced = False

            if isinstance(instr, SetAddrInstr):
                copy = instructions[0]
                assert isinstance(copy, CopyInstr)

                if len(instructions) >= 2 and isinstance(instructions[1], SetAddrInstr):
                    # ADDR, COPY, ADRR
                    if instr.shift == -instructions[1].shift:
                        # Replace with a write instead
                        merged.append(WriteInstr(old[instr.new : instr.new + copy.length]))
                        replaced = True
                        instructions.pop(0)
                        instructions.pop(0)
                elif (
                    len(instructions) >= 3
                    and isinstance(instructions[1], WriteInstr)
                    and isinstance(instructions[2], SetAddrInstr)
                    # ADDR, COPY, WRITE, ADRR
                    and instr.shift == -instructions[2].shift
                ):
                    write = instructions[1]
                    # Replace with a merged write instead
                    merged.append(WriteInstr(old[instr.new : instr.new + copy.length] + write.data))
                    replaced = True
                    instructions.pop(0)
                    instructions.pop(0)
                    instructions.pop(0)

            if not replaced:
                merged.append(instr)

        # We may have sequential WRITE commands due to the merging, do a pass
        cleaned = [merged[0]]
        for instr in merged[1:]:
            if isinstance(instr, WriteInstr) and isinstance(cleaned[-1], WriteInstr):
                cleaned[-1].data += instr.data
            else:
                cleaned.append(instr)
        return cleaned

    @classmethod
    def _merge_operations(cls, instructions: list[Instr]) -> list[Instr]:
        """Merge runs of COPY and WRITE into PATCH"""
        merged: list[Instr] = []
        to_merge: list[Instr] = []

        def finalise():
            nonlocal merged
            nonlocal to_merge
            if len(to_merge) == 0:
                return
            elif len(to_merge) == 1:
                merged.append(to_merge[0])
            else:
                merged.append(PatchInstr(to_merge))
            to_merge = []

        for instr in instructions:
            if (isinstance(instr, CopyInstr) and instr.length < 128) or (
                isinstance(instr, WriteInstr) and len(to_merge) > 0 and len(instr.data) < 256
            ):
                to_merge.append(instr)
            else:
                finalise()
                merged.append(instr)

        if len(to_merge) > 0:
            finalise()
        return merged

    @classmethod
    def _write_crack(cls, old: bytes, instructions: list[Instr]) -> list[Instr]:
        """Crack a WRITE operation into a [WRITE,COPY,WRITE] if COPY is at least 2 bytes"""

        cracked: list[Instr] = []
        old_offset = 0

        while len(instructions):
            instr = instructions.pop(0)

            if isinstance(instr, CopyInstr):
                old_offset = instr.original_offset + instr.length
                cracked.append(instr)
                continue
            elif isinstance(instr, SetAddrInstr):
                old_offset = instr.new
                cracked.append(instr)
                continue
            assert isinstance(instr, WriteInstr)

            split = [0]
            for idx, b in enumerate(instr.data):
                if old_offset + idx >= len(old):
                    # Add remainder of write to last split
                    split[-1] += len(instr.data) - idx
                    break
                if old[old_offset + idx] != b:
                    if len(split) % 2:
                        # Already on a WRITE
                        split[-1] += 1
                    else:
                        # On a COPY, swap to a WRITE
                        split.append(1)
                    continue

                if len(split) % 2:
                    # On a WRITE, switch to a COPY
                    split.append(1)
                else:
                    # Already on a COPY
                    split[-1] += 1

            # Total data count should remain the same
            assert sum(split) == len(instr.data)

            if len(split) % 2 == 0:
                # Ended on a copy
                copy_len = split.pop()
                if len(instructions) > 0 and isinstance(instructions[0], CopyInstr):
                    # Push the match into the next instruction if possible
                    instructions[0].length += copy_len
                    instructions[0].original_offset -= copy_len
                else:
                    # Merge the copy back into the previous write
                    split[-1] += copy_len

            # Should now have N*[WRITE, COPY] + [WRITE]
            assert len(split) % 2 == 1

            # Construct the [WRITE, COPY] pairs
            offset = 0
            while len(split) > 1:
                write_len = split.pop(0)
                copy_len = split.pop(0)

                # If the copy was only 1 byte, roll it back
                if copy_len == 1:
                    split[0] += write_len + copy_len
                else:
                    cracked.append(WriteInstr(instr.data[offset : offset + write_len]))
                    offset += write_len
                    cracked.append(CopyInstr(copy_len, old_offset + offset))
                    offset += copy_len

            # Append the final WRITE
            write_len = split.pop()
            cracked.append(WriteInstr(instr.data[offset : offset + write_len]))

        return cracked

    @classmethod
    def _gen_patch_instr(cls, bin_orig: bytes, bin_new: bytes) -> tuple[dict, list[Instr]]:
        best_patch = []
        best_patch_len = 2**32

        # Find best diff across range
        for i in range(4, 8):
            instr = cls._naive_diff(bin_orig, bin_new, i)
            instr = cls._cleanup_jumps(bin_orig, instr)
            instr = cls._write_crack(bin_orig, instr)
            instr = cls._merge_operations(instr)
            patch_len = sum([len(i) for i in instr])

            if patch_len < best_patch_len:
                best_patch = instr
                best_patch_len = patch_len

        metadata = {
            "original": {
                "len": len(bin_orig),
                "crc": binascii.crc32(bin_orig),
            },
            "new": {
                "len": len(bin_new),
                "crc": binascii.crc32(bin_new),
            },
        }

        return metadata, best_patch

    @classmethod
    def _gen_patch_header(cls, patch_metadata: dict, patch_data: bytes):
        hdr = cls.PatchHeader(
            cls.PatchHeader.magic_value,
            cls.PatchHeader.VERSION_MAJOR,
            cls.PatchHeader.VERSION_MINOR,
            cls.PatchHeader.ArrayValidation(
                patch_metadata["original"]["len"],
                patch_metadata["original"]["crc"],
            ),
            cls.PatchHeader.ArrayValidation(
                patch_metadata["new"]["len"],
                patch_metadata["new"]["crc"],
            ),
            cls.PatchHeader.ArrayValidation(
                len(patch_data),
                binascii.crc32(patch_data),
            ),
            0,
        )
        hdr_no_crc = bytes(hdr)
        hdr.header_crc = binascii.crc32(hdr_no_crc[: -ctypes.sizeof(ctypes.c_uint32)])
        return bytes(hdr)

    @classmethod
    def _gen_patch_data(cls, instructions: list[Instr]):
        output_bytes = b""
        for instr in instructions:
            output_bytes += bytes(instr)
        return output_bytes

    @classmethod
    def _patch_load(cls, patch_binary: bytes):
        hdr = cls.PatchHeader.from_buffer_copy(patch_binary)
        data = patch_binary[ctypes.sizeof(cls.PatchHeader) :]

        metadata = {
            "original": {
                "len": hdr.original_file.length,
                "crc": hdr.original_file.crc,
            },
            "new": {
                "len": hdr.constructed_file.length,
                "crc": hdr.constructed_file.crc,
            },
            "patch": {
                "len": hdr.patch_file.length,
                "crc": hdr.patch_file.crc,
            },
        }

        header_crc = binascii.crc32(patch_binary[: ctypes.sizeof(hdr) - ctypes.sizeof(ctypes.c_uint32)])
        if header_crc != hdr.header_crc:
            raise ValidationError("Patch header validation failed")
        if len(data) != hdr.patch_file.length:
            raise ValidationError(
                f"Patch data length does not match header information ({len(data)} != {hdr.patch_file.length})"
            )
        crc_patch = binascii.crc32(data)
        crc_expected = hdr.patch_file.crc
        if crc_patch != hdr.patch_file.crc:
            raise ValidationError(
                f"Patch data CRC does not match patch information ({crc_patch:08x} != {crc_expected:08x})"
            )

        instructions = []
        patch_offset = 0
        original_offset = 0
        while patch_offset < len(data):
            instr, length, original_offset = Instr.from_bytes(data, patch_offset, original_offset)
            patch_offset += length
            instructions.append(instr)

        return metadata, instructions

    @classmethod
    def generate(
        cls,
        bin_original: bytes,
        bin_new: bytes,
        verbose: bool,
    ) -> bytes:
        meta, instructions = cls._gen_patch_instr(bin_original, bin_new)
        patch_data = cls._gen_patch_data(instructions)
        patch_header = cls._gen_patch_header(meta, patch_data)
        bin_patch = patch_header + patch_data
        ratio = 100 * len(bin_patch) / meta["new"]["len"]

        print(f"Original File: {meta['original']['len']:6d} bytes")
        print(f"     New File: {meta['new']['len']:6d} bytes")
        print(f"   Patch File: {len(bin_patch):6d} bytes ({ratio:.2f}%) ({len(instructions):5d} instructions)")

        if verbose:
            class_count: dict[OpCode, int] = defaultdict(int)
            for instr in instructions:
                class_count[instr.ctypes_class().op] += 1

            print("")
            print("Instructions:")
            for instr_cls, instr_count in sorted(class_count.items()):
                print(f"{instr_cls.name:>16s}: {instr_count}")

        # Validate that file can be reconstructed
        patched = cls.patch(bin_original, bin_patch)
        assert bin_new == patched

        # Return complete file
        return bin_patch

    @classmethod
    def validation(cls, bin_original: bytes, invalid_length: bool, invalid_crc: bool) -> bytes:
        assert len(bin_original) > 1024

        # Manually construct an instruction set that runs all instructions
        instructions: list[Instr] = []
        instructions.append(WriteInstr(bin_original[:8], cls_override=WriteInstr.WriteU4))
        instructions.append(WriteInstr(bin_original[8:16], cls_override=WriteInstr.WriteU12))
        instructions.append(SetAddrInstr(16, 8, cls_override=SetAddrInstr.ShiftAddrS8))
        instructions.append(WriteInstr(bin_original[16:128], cls_override=WriteInstr.WriteU20))
        instructions.append(SetAddrInstr(120, 200, cls_override=SetAddrInstr.ShiftAddrS16))
        instructions.append(WriteInstr(bin_original[128:256], cls_override=WriteInstr.WriteU32))
        instructions.append(SetAddrInstr(328, 256, cls_override=SetAddrInstr.SetAddrU32))
        instructions.append(CopyInstr(8, cls_override=CopyInstr.CopyU4))
        instructions.append(CopyInstr(8, cls_override=CopyInstr.CopyU12))
        instructions.append(CopyInstr(128 - 16, cls_override=CopyInstr.CopyU20))
        instructions.append(CopyInstr(128, cls_override=CopyInstr.CopyU32))
        instructions.append(
            PatchInstr(
                [
                    CopyInstr(15),
                    WriteInstr(bin_original[512 + 15 : 512 + 16]),
                    CopyInstr(14),
                    WriteInstr(bin_original[512 + 30 : 512 + 32]),
                ]
            )
        )
        instructions.append(CopyInstr(len(bin_original) - 544))

        meta, _ = cls._gen_patch_instr(bin_original, bin_original)
        if invalid_length:
            meta["new"]["len"] -= 1
        if invalid_crc:
            meta["new"]["crc"] -= 1

        patch_data = cls._gen_patch_data(instructions)
        patch_header = cls._gen_patch_header(meta, patch_data)
        bin_patch = patch_header + patch_data

        # Validate that file can be reconstructed
        if not invalid_length and not invalid_crc:
            patched = cls.patch(bin_original, bin_patch)
            assert bin_original == patched

        return bin_patch

    @classmethod
    def patch(
        cls,
        bin_original: bytes,
        bin_patch: bytes,
    ) -> bytes:
        meta, instructions = cls._patch_load(bin_patch)
        patched = b""
        orig_offset = 0

        len_orig = len(bin_original)
        len_expected = meta["original"]["len"]
        if len_orig != len_expected:
            raise ValidationError(
                f"Original file length does not match patch information ({len_orig} != {len_expected})"
            )
        crc_orig = binascii.crc32(bin_original)
        crc_expected = meta["original"]["crc"]
        if crc_orig != crc_expected:
            raise ValidationError(
                f"Original file CRC does not match patch information ({crc_orig:08x} != {crc_expected:08x})"
            )

        for instr in instructions:
            if isinstance(instr, CopyInstr):
                patched += bin_original[orig_offset : orig_offset + instr.length]
                orig_offset += instr.length
            elif isinstance(instr, WriteInstr):
                patched += instr.data
                orig_offset += len(instr.data)
            elif isinstance(instr, SetAddrInstr):
                orig_offset = instr.new
            elif isinstance(instr, PatchInstr):
                for op in instr.operations:
                    if isinstance(op, CopyInstr):
                        patched += bin_original[orig_offset : orig_offset + op.length]
                        orig_offset += op.length
                    elif isinstance(op, WriteInstr):
                        patched += op.data
                        orig_offset += len(op.data)
                    else:
                        assert 0
            else:
                assert 0

        # Validate generated file matches what was expected
        len_patched = len(patched)
        len_expected = meta["new"]["len"]
        if len_patched != len_expected:
            raise ValidationError(
                f"Original file length does not match patch information ({len_patched} != {len_expected})"
            )
        crc_patched = binascii.crc32(patched)
        crc_expected = meta["new"]["crc"]
        if crc_patched != crc_expected:
            raise ValidationError(
                f"Original file CRC does not match patch information ({crc_patched:08x} != {crc_expected:08x})"
            )

        return patched

    @classmethod
    def dump(
        cls,
        bin_patch: bytes,
    ):
        meta, instructions = cls._patch_load(bin_patch)
        total_write_bytes = 0

        print(f"Original File: {meta['original']['len']:6d} bytes")
        print(f"     New File: {meta['new']['len']:6d} bytes")
        print(f"   Patch File: {len(bin_patch)} bytes ({len(instructions):5d} instructions)")

        class_count: dict[OpCode, int] = defaultdict(int)
        for instr in instructions:
            class_count[instr.ctypes_class().op] += 1
            if isinstance(instr, WriteInstr):
                total_write_bytes += len(instr.data)
            elif isinstance(instr, PatchInstr):
                for op in instr.operations:
                    if isinstance(op, WriteInstr):
                        total_write_bytes += len(op.data)

        print("")
        print("Total WRITE data:")
        print(f"\t{total_write_bytes} bytes ({100 * total_write_bytes / len(bin_patch):.2f}%)")

        print("")
        print("Instruction Count:")
        for op_cls, count in sorted(class_count.items()):
            print(f"{op_cls.name:>16s}: {count}")

        print("")
        print("Instruction List:")
        for instr in instructions:
            print(instr)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", "-v", action="store_true", help="Verbose command output")
    subparser = parser.add_subparsers(dest="command", title="Commands", required=True)

    # Generate patch file
    generate_args = subparser.add_parser("generate", help="Generate a patch file")
    generate_args.add_argument("original", help="Original file to use as base image")
    generate_args.add_argument("new", help="New file that will be the result of applying the patch")
    generate_args.add_argument("patch", help="Output patch file name")

    # Generate validation patch file
    validation_args = subparser.add_parser("validation", help="Generate a patch file for validating appliers")
    validation_args.add_argument("--invalid-length", action="store_true", help="Incorrect output file length")
    validation_args.add_argument("--invalid-crc", action="store_true", help="Incorrect output file CRC")
    validation_args.add_argument("input_file", help="File to use as base image and desired output")
    validation_args.add_argument("patch", help="Output patch file name")

    # Apply patch file
    patch_args = subparser.add_parser("patch", help="Apply a patch file")
    patch_args.add_argument("original", help="Original file to use as base image")
    patch_args.add_argument("patch", help="Patch file to apply")
    patch_args.add_argument("output", help="File to write output to")

    # Dump patch instructions
    dump_args = subparser.add_parser("dump", help="Dump patch file instructions to terminal")
    dump_args.add_argument("patch", help="Patch file to dump")

    # Parse args
    args = parser.parse_args()

    # Run requested command
    if args.command == "generate":
        with open(args.original, "rb") as f_orig, open(args.new, "rb") as f_new:
            patch = cpatch.generate(
                f_orig.read(-1),
                f_new.read(-1),
                args.verbose,
            )
        with open(args.patch, "wb") as f_output:
            f_output.write(patch)
    elif args.command == "validation":
        with open(args.input_file, "rb") as f_input:
            patch = cpatch.validation(f_input.read(-1), args.invalid_length, args.invalid_crc)
        with open(args.patch, "wb") as f_output:
            f_output.write(patch)
    elif args.command == "patch":
        with open(args.original, "rb") as f_orig, open(args.patch, "rb") as f_patch:
            output = cpatch.patch(f_orig.read(-1), f_patch.read(-1))
        with open(args.output, "wb") as f_output:
            f_output.write(output)
    elif args.command == "dump":
        with open(args.patch, "rb") as f_patch:
            cpatch.dump(f_patch.read(-1))



#!/usr/bin/env python3

import ctypes
import ipaddress

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr import net_if as z_nif
from infuse_iot.zephyr import wifi as z_wifi
from infuse_iot.zephyr.errno import errno


class interface_state(ctypes.LittleEndianStructure):
    _fields_ = [
        ("_state", ctypes.c_uint8),
        ("_if_flags", ctypes.c_uint32),
        ("_l2_flags", ctypes.c_uint16),
        ("mtu", ctypes.c_uint16),
        ("ipv4", 4 * ctypes.c_uint8),
        ("ipv6", 16 * ctypes.c_uint8),
    ]
    _pack_ = 1

    @property
    def state(self):
        return z_nif.OperationalState(self._state)

    @property
    def if_flags(self):
        return z_nif.InterfaceFlags(self._if_flags)

    @property
    def l2_flags(self):
        return z_nif.L2Flags(self._l2_flags)


class wifi_state_struct(ctypes.LittleEndianStructure):
    _fields_ = [
        ("_state", ctypes.c_uint8),
        ("ssid", 32 * ctypes.c_char),
        ("bssid", 6 * ctypes.c_char),
        ("_band", ctypes.c_uint8),
        ("channel", ctypes.c_uint8),
        ("_iface_mode", ctypes.c_uint8),
        ("_link_mode", ctypes.c_uint8),
        ("_security", ctypes.c_uint8),
        ("rssi", ctypes.c_int8),
        ("beacon_interval", ctypes.c_uint16),
        ("twt_capable", ctypes.c_uint8),
    ]
    _pack_ = 1

    @property
    def state(self):
        return z_wifi.WiFiState(self._state)

    @property
    def band(self):
        return z_wifi.FrequencyBand(self._band)

    @property
    def interface_mode(self):
        return z_wifi.InterfaceMode(self._iface_mode)

    @property
    def link_mode(self):
        return z_wifi.LinkMode(self._link_mode)

    @property
    def security_type(self):
        return z_wifi.SecurityType(self._security)


class wifi_state(InfuseRpcCommand, defs.wifi_state):
    class response(ctypes.LittleEndianStructure):
        _fields_ = [
            ("common", interface_state),
            ("wifi", wifi_state_struct),
        ]
        _pack_ = 1

    @classmethod
    def add_parser(cls, parser):
        return

    def __init__(self, args):
        self.args = args

    def request_struct(self):
        return self.request()

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query current time ({errno.strerror(-return_code)})")
            return

        common = response.common
        wifi = response.wifi

        # Address formatting
        ipv4 = ipaddress.IPv4Address(bytes(common.ipv4))
        ipv6 = ipaddress.IPv6Address(bytes(common.ipv6))
        bssid = ":".join([f"{b:02x}" for b in wifi.bssid])

        print("Interface State:")
        print(f"\t          State: {common.state.name}")
        print(f"\t       IF Flags: {common.if_flags}")
        print(f"\t       L2 Flags: {common.l2_flags}")
        print(f"\t            MTU: {common.mtu}")
        print(f"\t           IPv4: {ipv4}")
        print(f"\t           IPv6: {ipv6}")
        print("WiFi State:")
        print(f"\t          State: {wifi.state.name}")
        if wifi.state >= z_wifi.WiFiState.AUTHENTICATING:
            try:
                print(f"\t           SSID: {wifi.ssid.decode('utf-8')}")
            except UnicodeDecodeError:
                print(f"\t           SSID: Decode Error (Hex: {wifi.ssid.hex()})")
            print(f"\t          BSSID: {bssid}")
            print(f"\t Frequency Band: {wifi.band}")
            print(f"\t        Channel: {wifi.channel}")
            print(f"\t        IF Mode: {wifi.interface_mode.name}")
            print(f"\t      Link Mode: {wifi.link_mode}")
            print(f"\t       Security: {wifi.security_type}")
            print(f"\t           RSSI: {wifi.rssi} dBm")
            print(f"\tBeacon Interval: {wifi.beacon_interval} ms")
            print(f"\t    TWT Capable: {'Yes' if wifi.twt_capable else 'No'}")



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class reboot(InfuseRpcCommand, defs.reboot):
    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--delay", type=int, default=0, help="Delay until reboot (ms)")

    def __init__(self, args):
        self._delay_ms = args.delay

    def request_struct(self):
        return self.request(self._delay_ms)

    def request_json(self):
        return {"delay_ms": str(self._delay_ms)}

    def handle_response(self, return_code, response):
        if return_code == 0:
            print(f"Rebooting in {response.delay_ms} ms")
        else:
            print(f"Failed to trigger reboot ({errno.strerror(-return_code)})")



#!/usr/bin/env python3

import base64
import binascii

from infuse_iot.api_client import Client
from infuse_iot.api_client.api.key import get_shared_secret
from infuse_iot.api_client.models import Key
from infuse_iot.credentials import get_api_key, load_network
from infuse_iot.epacket.interface import Address as InterfaceAddress
from infuse_iot.util.crypto import hkdf_derive


class NoKeyError(KeyError):
    """Generic key not found error"""


class UnknownNetworkError(NoKeyError):
    """Requested network is not known"""


class DeviceUnknownDeviceKey(NoKeyError):
    """Device key is not known for requested device"""


class DeviceUnknownNetworkKey(NoKeyError):
    """Network key is not known for requested device"""


class DeviceKeyChangedError(KeyError):
    """Device key for the requested device has changed"""


class DeviceDatabase:
    """Database of current device state"""

    _network_keys: dict[int, bytes] = {}
    _derived_keys: dict[tuple[int, bytes, int], bytes] = {}

    class DeviceState:
        """Device State"""

        def __init__(
            self,
            address: int,
            network_id: int | None = None,
            device_id: int | None = None,
        ):
            self.address = address
            self.network_id = network_id
            self.device_id = device_id
            self.bt_addr: InterfaceAddress.BluetoothLeAddr | None = None
            self.public_key: bytes | None = None
            self.shared_key: bytes | None = None
            self._tx_gatt_seq = 0

        def gatt_sequence_num(self):
            """Persistent auto-incrementing sequence number for GATT"""
            self._tx_gatt_seq += 1
            return self._tx_gatt_seq

    def __init__(self) -> None:
        self.gateway: int | None = None
        self.devices: dict[int, DeviceDatabase.DeviceState] = {}
        self.bt_addr: dict[InterfaceAddress.BluetoothLeAddr, int] = {}

    def observe_device(
        self,
        address: int,
        network_id: int | None = None,
        device_id: int | None = None,
        bt_addr: InterfaceAddress.BluetoothLeAddr | None = None,
    ) -> None:
        """Update device state based on observed packet"""
        if self.gateway is None:
            self.gateway = address
        if address not in self.devices:
            self.devices[address] = self.DeviceState(address)
        if network_id is not None:
            self.devices[address].network_id = network_id
        if device_id is not None:
            if self.devices[address].device_id is not None and self.devices[address].device_id != device_id:
                raise DeviceKeyChangedError(f"Device key for {address:016x} has changed")
            self.devices[address].device_id = device_id
        if bt_addr is not None:
            self.bt_addr[bt_addr] = address
            self.devices[address].bt_addr = bt_addr

    def observe_security_state(self, address: int, cloud_key: bytes, device_key: bytes, network_id: int) -> None:
        """Update device state based on security_state response"""
        if address not in self.devices:
            self.devices[address] = self.DeviceState(address)
        device_id = binascii.crc32(cloud_key + device_key) & 0x00FFFFFF
        self.devices[address].device_id = device_id
        self.devices[address].network_id = network_id
        self.devices[address].public_key = device_key

        client = Client(base_url="https://api.infuse-iot.com").with_headers({"x-api-key": f"Bearer {get_api_key()}"})

        with client as client:
            body = Key(base64.b64encode(device_key).decode("utf-8"))
            response = get_shared_secret.sync(client=client, body=body)
            if response is not None:
                key = base64.b64decode(response.key)
                self.devices[address].shared_key = key

    def _network_key(self, network_id: int, interface: bytes, gps_time: int) -> bytes:
        if network_id not in self._network_keys:
            try:
                info = load_network(network_id)
            except FileNotFoundError:
                raise UnknownNetworkError(network_id) from None
            self._network_keys[network_id] = info["key"]
        base = self._network_keys[network_id]
        time_idx = gps_time // (60 * 60 * 24)

        key_id = (network_id, interface, time_idx)
        if key_id not in self._derived_keys:
            self._derived_keys[key_id] = hkdf_derive(base, time_idx.to_bytes(4, "little"), interface)

        return self._derived_keys[key_id]

    def _serial_key(self, base: bytes, time_idx: int) -> bytes:
        return hkdf_derive(base, time_idx.to_bytes(4, "little"), b"serial")

    def _bt_adv_key(self, base: bytes, time_idx: int) -> bytes:
        return hkdf_derive(base, time_idx.to_bytes(4, "little"), b"bt_adv")

    def _bt_gatt_key(self, base: bytes, time_idx: int) -> bytes:
        return hkdf_derive(base, time_idx.to_bytes(4, "little"), b"bt_gatt")

    def _udp_key(self, base: bytes, time_idx: int) -> bytes:
        return hkdf_derive(base, time_idx.to_bytes(4, "little"), b"udp")

    def has_public_key(self, address: int) -> bool:
        """Does the database have the public key for this device?"""
        if address not in self.devices:
            return False
        return self.devices[address].public_key is not None

    def has_network_id(self, address: int) -> bool:
        """Does the database know the network ID for this device?"""
        if address not in self.devices:
            return False
        return self.devices[address].network_id is not None

    def infuse_id_from_bluetooth(self, bt_addr: InterfaceAddress.BluetoothLeAddr) -> int | None:
        """Get Bluetooth address associated with device"""
        return self.bt_addr.get(bt_addr, None)

    def serial_network_key(self, address: int, gps_time: int) -> bytes:
        """Network key for serial interface"""
        if address not in self.devices:
            raise DeviceUnknownNetworkKey
        network_id = self.devices[address].network_id
        if network_id is None:
            raise DeviceUnknownNetworkKey

        return self._network_key(network_id, b"serial", gps_time)

    def serial_device_key(self, address: int, gps_time: int) -> bytes:
        """Device key for serial interface"""
        if address not in self.devices:
            raise DeviceUnknownDeviceKey
        d = self.devices[address]
        if d.device_id is None:
            raise DeviceUnknownDeviceKey
        base = self.devices[address].shared_key
        if base is None:
            raise DeviceUnknownDeviceKey
        time_idx = gps_time // (60 * 60 * 24)

        return self._serial_key(base, time_idx)

    def bt_adv_network_key(self, address: int, gps_time: int) -> bytes:
        """Network key for Bluetooth advertising interface"""
        if address not in self.devices:
            raise DeviceUnknownNetworkKey
        network_id = self.devices[address].network_id
        if network_id is None:
            raise DeviceUnknownNetworkKey

        return self._network_key(network_id, b"bt_adv", gps_time)

    def bt_adv_device_key(self, address: int, gps_time: int) -> bytes:
        """Device key for Bluetooth advertising interface"""
        if address not in self.devices:
            raise DeviceUnknownDeviceKey
        d = self.devices[address]
        if d.device_id is None:
            raise DeviceUnknownDeviceKey
        base = self.devices[address].shared_key
        if base is None:
            raise DeviceUnknownDeviceKey
        time_idx = gps_time // (60 * 60 * 24)

        return self._bt_adv_key(base, time_idx)

    def bt_gatt_network_key(self, address: int, gps_time: int) -> bytes:
        """Network key for Bluetooth advertising interface"""
        if address not in self.devices:
            raise DeviceUnknownNetworkKey
        network_id = self.devices[address].network_id
        if network_id is None:
            raise DeviceUnknownNetworkKey

        return self._network_key(network_id, b"bt_gatt", gps_time)

    def bt_gatt_device_key(self, address: int, gps_time: int) -> bytes:
        """Device key for Bluetooth advertising interface"""
        if address not in self.devices:
            raise DeviceUnknownDeviceKey
        d = self.devices[address]
        if d.device_id is None:
            raise DeviceUnknownDeviceKey
        base = self.devices[address].shared_key
        if base is None:
            raise DeviceUnknownDeviceKey
        time_idx = gps_time // (60 * 60 * 24)

        return self._bt_gatt_key(base, time_idx)

    def udp_network_key(self, address: int, gps_time: int) -> bytes:
        """Network key for UDP interface"""
        if address not in self.devices:
            raise DeviceUnknownNetworkKey
        network_id = self.devices[address].network_id
        if network_id is None:
            raise DeviceUnknownNetworkKey

        return self._network_key(network_id, b"udp", gps_time)

    def udp_device_key(self, address: int, gps_time: int) -> bytes:
        """Device key for UDP interface"""
        if address not in self.devices:
            raise DeviceUnknownDeviceKey
        d = self.devices[address]
        if d.device_id is None:
            raise DeviceUnknownDeviceKey
        base = self.devices[address].shared_key
        if base is None:
            raise DeviceUnknownDeviceKey
        time_idx = gps_time // (60 * 60 * 24)

        return self._udp_key(base, time_idx)



#!/usr/bin/env python3

import enum


class errno(enum.IntEnum):
    """Error codes exported from Zephyr `errno.h`"""

    EPERM = (1, "Not owner")
    ENOENT = (2, "No such file or directory")
    ESRCH = (3, "No such context")
    EINTR = (4, "Interrupted system call")
    EIO = (5, "I/O error")
    ENXIO = (6, "No such device or address")
    E2BIG = (7, "Arg list too long")
    ENOEXEC = (8, "Exec format error")
    EBADF = (9, "Bad file number")
    ECHILD = (10, "No children")
    EAGAIN = (11, "No more contexts")
    ENOMEM = (12, "Not enough core")
    EACCES = (13, "Permission denied")
    EFAULT = (14, "Bad address")
    ENOTBLK = (15, "Block device required")
    EBUSY = (16, "Mount device busy")
    EEXIST = (17, "File exists")
    EXDEV = (18, "Cross-device link")
    ENODEV = (19, "No such device")
    ENOTDIR = (20, "Not a directory")
    EISDIR = (21, "Is a directory")
    EINVAL = (22, "Invalid argument")
    ENFILE = (23, "File table overflow")
    EMFILE = (24, "Too many open files")
    ENOTTY = (25, "Not a typewriter")
    ETXTBSY = (26, "Text file busy")
    EFBIG = (27, "File too large")
    ENOSPC = (28, "No space left on device")
    ESPIPE = (29, "Illegal seek")
    EROFS = (30, "Read-only file system")
    EMLINK = (31, "Too many links")
    EPIPE = (32, "Broken pipe")
    EDOM = (33, "Argument too large")
    ERANGE = (34, "Result too large")
    ENOMSG = (35, "Unexpected message type")
    EDEADLK = (45, "Resource deadlock avoided")
    ENOLCK = (46, "No locks available")
    ENOSTR = (60, "STREAMS device required")
    ENODATA = (61, "Missing expected message data")
    ETIME = (62, "STREAMS timeout occurred")
    ENOSR = (63, "Insufficient memory")
    EPROTO = (71, "Generic STREAMS error")
    EBADMSG = (77, "Invalid STREAMS message")
    ENOSYS = (88, "Function not implemented")
    ENOTEMPTY = (90, "Directory not empty")
    ENAMETOOLONG = (91, "File name too long")
    ELOOP = (92, "Too many levels of symbolic links")
    EOPNOTSUPP = (95, "Operation not supported on socket")
    EPFNOSUPPORT = (96, "Protocol family not supported")
    ECONNRESET = (104, "Connection reset by peer")
    ENOBUFS = (105, "No buffer space available")
    EAFNOSUPPORT = (106, "Addr family not supported")
    EPROTOTYPE = (107, "Protocol wrong type for socket")
    ENOTSOCK = (108, "Socket operation on non-socket")
    ENOPROTOOPT = (109, "Protocol not available")
    ESHUTDOWN = (110, "Can't send after socket shutdown")
    ECONNREFUSED = (111, "Connection refused")
    EADDRINUSE = (112, "Address already in use")
    ECONNABORTED = (113, "Software caused connection abort")
    ENETUNREACH = (114, "Network is unreachable")
    ENETDOWN = (115, "Network is down")
    ETIMEDOUT = (116, "Connection timed out")
    EHOSTDOWN = (117, "Host is down")
    EHOSTUNREACH = (118, "No route to host")
    EINPROGRESS = (119, "Operation now in progress")
    EALREADY = (120, "Operation already in progress")
    EDESTADDRREQ = (121, "Destination address required")
    EMSGSIZE = (122, "Message size")
    EPROTONOSUPPORT = (123, "Protocol not supported")
    ESOCKTNOSUPPORT = (124, "Socket type not supported")
    EADDRNOTAVAIL = (125, "Can't assign requested address")
    ENETRESET = (126, "Network dropped connection on reset")
    EISCONN = (127, "Socket is already connected")
    ENOTCONN = (128, "Socket is not connected")
    ETOOMANYREFS = (129, "Too many references: can't splice")
    ENOTSUP = (134, "Unsupported value/command")
    EILSEQ = (138, "Illegal byte sequence")
    EOVERFLOW = (139, "Value overflow")
    ECANCELED = (140, "Operation canceled")

    description: str

    def __new__(cls, value: int, description: str = ""):
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.description = description
        return obj

    @classmethod
    def strerror(cls, int) -> str:
        return cls(int).description



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.definitions_field_conversion import DefinitionsFieldConversion
    from ..models.definitions_field_display import DefinitionsFieldDisplay


T = TypeVar("T", bound="DefinitionsFieldDefinition")


@_attrs_define
class DefinitionsFieldDefinition:
    """
    Attributes:
        name (str): Field name
        type_ (str): Field type
        description (Union[Unset, str]): Field description
        num (Union[Unset, int]): If field is array, the number of elements (0 for variable length)
        counted_by (Union[Unset, str]): If field is array, the name of the field that contains the number of elements
            (overrides num)
        display (Union[Unset, DefinitionsFieldDisplay]): Display settings for a field
        conversion (Union[Unset, DefinitionsFieldConversion]): Conversion formula for a field (m * <value> + c)
    """

    name: str
    type_: str
    description: Unset | str = UNSET
    num: Unset | int = UNSET
    counted_by: Unset | str = UNSET
    display: Union[Unset, "DefinitionsFieldDisplay"] = UNSET
    conversion: Union[Unset, "DefinitionsFieldConversion"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        type_ = self.type_

        description = self.description

        num = self.num

        counted_by = self.counted_by

        display: Unset | dict[str, Any] = UNSET
        if not isinstance(self.display, Unset):
            display = self.display.to_dict()

        conversion: Unset | dict[str, Any] = UNSET
        if not isinstance(self.conversion, Unset):
            conversion = self.conversion.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "type": type_,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if num is not UNSET:
            field_dict["num"] = num
        if counted_by is not UNSET:
            field_dict["counted_by"] = counted_by
        if display is not UNSET:
            field_dict["display"] = display
        if conversion is not UNSET:
            field_dict["conversion"] = conversion

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.definitions_field_conversion import DefinitionsFieldConversion
        from ..models.definitions_field_display import DefinitionsFieldDisplay

        d = dict(src_dict)
        name = d.pop("name")

        type_ = d.pop("type")

        description = d.pop("description", UNSET)

        num = d.pop("num", UNSET)

        counted_by = d.pop("counted_by", UNSET)

        _display = d.pop("display", UNSET)
        display: Unset | DefinitionsFieldDisplay
        if isinstance(_display, Unset):
            display = UNSET
        else:
            display = DefinitionsFieldDisplay.from_dict(_display)

        _conversion = d.pop("conversion", UNSET)
        conversion: Unset | DefinitionsFieldConversion
        if isinstance(_conversion, Unset):
            conversion = UNSET
        else:
            conversion = DefinitionsFieldConversion.from_dict(_conversion)

        definitions_field_definition = cls(
            name=name,
            type_=type_,
            description=description,
            num=num,
            counted_by=counted_by,
            display=display,
            conversion=conversion,
        )

        definitions_field_definition.additional_properties = d
        return definitions_field_definition

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="ApplicationVersion")


@_attrs_define
class ApplicationVersion:
    """Application version

    Attributes:
        major (int): Major version number
        minor (int): Minor version number
        revision (int): Revision number
        build_num (int): Build number
    """

    major: int
    minor: int
    revision: int
    build_num: int
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        major = self.major

        minor = self.minor

        revision = self.revision

        build_num = self.build_num

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "major": major,
                "minor": minor,
                "revision": revision,
                "buildNum": build_num,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        major = d.pop("major")

        minor = d.pop("minor")

        revision = d.pop("revision")

        build_num = d.pop("buildNum")

        application_version = cls(
            major=major,
            minor=minor,
            revision=revision,
            build_num=build_num,
        )

        application_version.additional_properties = d
        return application_version

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

import ctypes

import infuse_iot.definitions.rpc as defs
import infuse_iot.zephyr.wifi as wifi
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.rpc_wrappers.kv_write import kv_write
from infuse_iot.util.ctypes import UINT8_MAX, VLACompatLittleEndianStruct
from infuse_iot.zephyr.errno import errno


class wifi_configure(InfuseRpcCommand, defs.kv_write):
    HELP = "Set the WiFi network SSID and PSK"
    DESCRIPTION = "Set the WiFi network SSID and PSK"

    class request(ctypes.LittleEndianStructure):
        _fields_ = [
            ("num", ctypes.c_uint8),
        ]
        _pack_ = 1

    class response(VLACompatLittleEndianStruct):
        vla_field = ("rc", 0 * ctypes.c_int16)

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--ssid", "-s", type=str, help="Network name")
        parser.add_argument("--psk", "-p", type=str, help="Network password")
        band_parser = parser.add_mutually_exclusive_group()
        band_parser.add_argument(
            "--band-unknown",
            action="store_const",
            dest="band",
            const=UINT8_MAX,
            default=UINT8_MAX,
            help="Unknown frequency band",
        )
        band_parser.add_argument(
            "--band-2G4",
            action="store_const",
            dest="band",
            const=wifi.FrequencyBand.BAND_2_4_GHZ,
            help="2.4GHz frequency band",
        )
        band_parser.add_argument(
            "--band-5G",
            action="store_const",
            dest="band",
            const=wifi.FrequencyBand.BAND_5_GHZ,
            help="5GHz frequency band",
        )
        band_parser.add_argument(
            "--band-6G",
            action="store_const",
            dest="band",
            const=wifi.FrequencyBand.BAND_6_GHZ,
            help="6GHz frequency band",
        )
        parser.add_argument(
            "--channel", "-c", type=int, default=wifi.FrequencyChannel.CHANNEL_ANY, help="Network channel index"
        )

    def __init__(self, args):
        self.args = args

    def request_struct(self):
        ssid_bytes = self.args.ssid.encode("utf-8") + b"\x00"
        psk_bytes = self.args.psk.encode("utf-8") + b"\x00"
        chan_bytes = self.args.band.to_bytes(1, "little") + self.args.channel.to_bytes(1, "little")

        ssid_struct = kv_write.kv_store_value_factory(20, len(ssid_bytes).to_bytes(1, "little") + ssid_bytes)
        psk_struct = kv_write.kv_store_value_factory(21, len(psk_bytes).to_bytes(1, "little") + psk_bytes)
        chan_struct = kv_write.kv_store_value_factory(22, chan_bytes)

        request_bytes = bytes(ssid_struct) + bytes(psk_struct) + bytes(chan_struct)
        return bytes(self.request(3)) + request_bytes

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Invalid data buffer ({errno.strerror(-return_code)})")
            return

        def print_status(name, rc):
            if rc < 0:
                print(f"{name} failed to write")
            elif rc == 0:
                print(f"{name} already matched")
            else:
                print(f"{name} updated")

        print_status("SSID", response.rc[0])
        print_status("PSK", response.rc[1])
        print_status("Channel", response.rc[2])



#!/usr/bin/env python3


def humanised_seconds(seconds: int, units: int = 2) -> str:
    """Convert a seconds count to human readable units"""
    scales = [365 * 24 * 60 * 60, 7 * 24 * 60 * 60, 24 * 60 * 60, 60 * 60, 60, 1]
    postfix = ["year", "week", "day", "hour", "minute", "second"]

    vals = []
    for scale in scales:
        vals.append(seconds // scale)
        seconds -= scale * vals[-1]

    def construct(val, postfix):
        return f"{val} {postfix}{'' if val == 1 else 's'}"

    for idx, val in enumerate(vals):
        if val == 0:
            continue
        units = min(units, len(postfix) - idx)
        final = []
        while units > 0:
            final.append(construct(vals[idx], postfix[idx]))
            idx += 1
            units -= 1

        return ", ".join(final)
    return "0 seconds"



# Copyright (c) 2019, Nordic Semiconductor ASA
#
# Don't put anything else in here!
#
# This is the Python 3 version of option 3 in:
# https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version

import importlib.metadata

__version__ = importlib.metadata.version("infuse_iot")



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.new_device_kv_entry_update_decoded import NewDeviceKVEntryUpdateDecoded


T = TypeVar("T", bound="NewDeviceKVEntryUpdate")


@_attrs_define
class NewDeviceKVEntryUpdate:
    """
    Attributes:
        data (Union[Unset, str]): Raw entry data as a base64 encoded string (must provide either data or decoded)
        decoded (Union[Unset, NewDeviceKVEntryUpdateDecoded]): Decoded entry value (must provide either data or decoded)
    """

    data: Unset | str = UNSET
    decoded: Union[Unset, "NewDeviceKVEntryUpdateDecoded"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        data = self.data

        decoded: Unset | dict[str, Any] = UNSET
        if not isinstance(self.decoded, Unset):
            decoded = self.decoded.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if data is not UNSET:
            field_dict["data"] = data
        if decoded is not UNSET:
            field_dict["decoded"] = decoded

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.new_device_kv_entry_update_decoded import NewDeviceKVEntryUpdateDecoded

        d = dict(src_dict)
        data = d.pop("data", UNSET)

        _decoded = d.pop("decoded", UNSET)
        decoded: Unset | NewDeviceKVEntryUpdateDecoded
        if isinstance(_decoded, Unset):
            decoded = UNSET
        else:
            decoded = NewDeviceKVEntryUpdateDecoded.from_dict(_decoded)

        new_device_kv_entry_update = cls(
            data=data,
            decoded=decoded,
        )

        new_device_kv_entry_update.additional_properties = d
        return new_device_kv_entry_update

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

"""Manage Infuse-IoT credentials"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import yaml

from infuse_iot import credentials
from infuse_iot.commands import InfuseCommand
from infuse_iot.util.argparse import ValidDir, ValidFile


class SubCommand(InfuseCommand):
    NAME = "credentials"
    HELP = "Manage Infuse-IoT credentials"
    DESCRIPTION = "Manage Infuse-IoT credentials"

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--api-key", type=str, help="Set Infuse-IoT API key")
        parser.add_argument("--api-key-print", action="store_true", help="Print Infuse-IoT API key")
        parser.add_argument("--network", type=ValidFile, help="Load network credentials from file")
        parser.add_argument("--custom-tools", type=ValidDir, help="Location of custom tools")
        parser.add_argument("--custom-definitions", type=ValidDir, help="Location of custom definitions")

    def __init__(self, args):
        self.args = args

    def run(self):
        if self.args.api_key is not None:
            credentials.set_api_key(self.args.api_key)
        if self.args.api_key_print:
            try:
                print(f"API Key: {credentials.get_api_key()}")
            except FileNotFoundError:
                print("API Key: N/A")
        if self.args.network is not None:
            # Read the file
            with self.args.network.open("r") as f:
                content = f.read()
            # Validate it is valid yaml
            network_info = yaml.safe_load(content)
            credentials.save_network(network_info["id"], content)
        if self.args.custom_tools:
            credentials.set_custom_tool_path(str(self.args.custom_tools.absolute()))
        if self.args.custom_definitions:
            credentials.set_custom_definitions_path(str(self.args.custom_definitions.absolute()))



#!/usr/bin/env python3

"""Infuse-IoT cloud interaction"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import sys
from http import HTTPStatus
from json import loads
from typing import Any

from tabulate import tabulate

from infuse_iot.api_client import Client
from infuse_iot.api_client.api.board import (
    create_board,
    get_board_by_id,
    get_boards,
)
from infuse_iot.api_client.api.coap import get_coap_files
from infuse_iot.api_client.api.device import (
    get_device_by_device_id,
    get_device_kv_entries_by_device_id,
    get_device_last_route_by_device_id,
    get_device_state_by_id,
)
from infuse_iot.api_client.api.organisation import (
    create_organisation,
    get_all_organisations,
    get_organisation_by_id,
)
from infuse_iot.api_client.models import COAPFilesList, Error, NewBoard, NewOrganisation
from infuse_iot.api_client.types import Unset
from infuse_iot.commands import InfuseCommand
from infuse_iot.credentials import get_api_key


class CloudSubCommand:
    def __init__(self, args):
        self.args = args

    def run(self):
        """Run cloud sub-command"""

    def client(self):
        """Get API client object ready to use"""
        return Client(base_url="https://api.infuse-iot.com").with_headers({"x-api-key": f"Bearer {get_api_key()}"})


class Organisations(CloudSubCommand):
    @classmethod
    def add_parser(cls, parser):
        parser_orgs = parser.add_parser("orgs", help="Infuse-IoT organisations")
        parser_orgs.set_defaults(command_class=cls)

        tool_parser = parser_orgs.add_subparsers(title="commands", metavar="<command>", required=True)

        list_parser = tool_parser.add_parser("list")
        list_parser.set_defaults(command_fn=cls.list)

        create_parser = tool_parser.add_parser("create")
        create_parser.add_argument("--name", "-n", type=str, required=True)
        create_parser.set_defaults(command_fn=cls.create)

    def run(self):
        with self.client() as client:
            self.args.command_fn(self, client)

    def list(self, client: Client):
        org_list = []

        orgs = get_all_organisations.sync(client=client)
        if isinstance(orgs, Error) or orgs is None:
            sys.exit(f"Organisation query failed {orgs}")
        for o in orgs:
            org_list.append([o.name, o.id])

        print(
            tabulate(
                org_list,
                headers=["Name", "ID"],
            )
        )

    def create(self, client: Client):
        rsp = create_organisation.sync_detailed(
            client=client,
            body=NewOrganisation(self.args.name),
        )

        if rsp.status_code == HTTPStatus.CREATED:
            assert rsp.parsed is not None
            print(f"Created organisation {rsp.parsed.name} with ID {rsp.parsed.id}")
        else:
            c = loads(rsp.content.decode("utf-8"))
            print(f"<{rsp.status_code}>: {c['message']}")


class Boards(CloudSubCommand):
    @classmethod
    def add_parser(cls, parser):
        parser_boards = parser.add_parser("boards", help="Infuse-IoT hardware platforms")
        parser_boards.set_defaults(command_class=cls)

        tool_parser = parser_boards.add_subparsers(title="commands", metavar="<command>", required=True)

        list_parser = tool_parser.add_parser("list")
        list_parser.set_defaults(command_fn=cls.list)

        create_parser = tool_parser.add_parser("create")
        create_parser.add_argument("--name", "-n", type=str, required=True, help="New board name")
        create_parser.add_argument("--org", "-o", type=str, required=True, help="Organisation ID")
        create_parser.add_argument("--soc", "-s", type=str, required=True, help="Board system on chip")
        create_parser.add_argument("--desc", "-d", type=str, required=True, help="Board description")
        create_parser.set_defaults(command_fn=cls.create)

    def run(self):
        with self.client() as client:
            self.args.command_fn(self, client)

    def list(self, client: Client):
        board_list = []

        orgs = get_all_organisations.sync(client=client)
        if isinstance(orgs, Error) or orgs is None:
            sys.exit(f"Organisation query failed {orgs}")
        for org in orgs:
            boards = get_boards.sync(client=client, organisation_id=org.id)
            if isinstance(boards, Error) or boards is None:
                sys.exit(f"Boards query failed {boards}")

            for b in boards:
                board_list.append([b.name, b.id, b.soc, org.name, b.description])

        print(
            tabulate(
                board_list,
                headers=["Name", "ID", "SoC", "Organisation", "Description"],
            )
        )

    def create(self, client: Client):
        rsp = create_board.sync_detailed(
            client=client,
            body=NewBoard(
                name=self.args.name,
                description=self.args.desc,
                soc=self.args.soc,
                organisation_id=self.args.org,
            ),
        )
        if rsp.status_code == HTTPStatus.CREATED:
            assert rsp.parsed is not None
            print(f"Created board {rsp.parsed.name} with ID {rsp.parsed.id}")
        else:
            c = loads(rsp.content.decode("utf-8"))
            print(f"<{rsp.status_code}>: {c['message']}")


class Device(CloudSubCommand):
    @classmethod
    def add_parser(cls, parser):
        parser_boards = parser.add_parser("device", help="Infuse-IoT devices")
        parser_boards.set_defaults(command_class=cls)

        tool_parser = parser_boards.add_subparsers(title="commands", metavar="<command>", required=True)

        info_parser = tool_parser.add_parser("info")
        info_parser.set_defaults(command_fn=cls.info)
        info_parser.add_argument("--id", type=str, help="Infuse-IoT device ID")
        info_parser = tool_parser.add_parser("kv_state")
        info_parser.set_defaults(command_fn=cls.kv_state)
        info_parser.add_argument("--id", type=str, help="Infuse-IoT device ID")

    def run(self):
        with self.client() as client:
            self.args.command_fn(self, client)

    def info(self, client: Client):
        id_int = int(self.args.id, 0)
        id_str = f"{id_int:016x}"
        info = get_device_by_device_id.sync(client=client, device_id=id_str)
        if info is None:
            sys.exit(f"No device with Infuse-IoT ID {id_str} found")
        metadata: list[tuple[str, Any]] = []
        if info.metadata:
            metadata = [(f"Metadata.{k}", v) for k, v in info.metadata.additional_properties.items()]

        org = get_organisation_by_id.sync(client=client, id=info.organisation_id)
        board = get_board_by_id.sync(client=client, id=info.board_id)
        state = get_device_state_by_id.sync(client=client, id=info.id)
        route = get_device_last_route_by_device_id.sync(client=client, device_id=id_str)

        table: list[tuple[str, Any]] = [
            ("UUID", info.id),
            ("MCU ID", info.mcu_id),
            ("Organisation", f"{info.organisation_id} ({org.name if org else 'Unknown'})"),
            ("Board", f"{info.board_id} ({board.name if board else 'Unknown'})"),
            ("Created", info.created_at),
            ("Updated", info.updated_at),
            *metadata,
        ]
        if state is not None:
            v = state.application_version

            table += [
                ("~~~State~~~", ""),
                ("Updated", state.updated_at),
            ]
            if state.application_id:
                table += [("Application ID", f"0x{state.application_id:08x}")]
            if v:
                table += [("Version", f"{v.major}.{v.minor}.{v.revision}+{v.build_num:08x}")]
        if route is not None:
            table += [
                ("~~~Latest Route~~~", ""),
                ("Interface", route.interface.upper()),
            ]
            if route.bt_adv:
                table += [("BT Address", f"{route.bt_adv.address} ({route.bt_adv.type_})")]

        print(tabulate(table))

    def _kv_display(self, table: list[tuple[str, Any]], name_base: str, dictionary: dict):
        for name, value in dictionary.items():
            if isinstance(value, dict):
                self._kv_display(table, f"{name_base}.{name}", value)
            else:
                table.append((f"{name_base}.{name}", value))

    def kv_state(self, client: Client):
        id_int = int(self.args.id, 0)
        id_str = f"{id_int:016x}"

        kv_state = get_device_kv_entries_by_device_id.sync(client=client, device_id=id_str)
        if kv_state is None:
            print(f"Unable to query KV state for {id_str}")
            return

        table: list[tuple[str, Any]] = []
        for element in kv_state:
            key = element.key_name if isinstance(element.key_name, str) else str(element.key_id)

            if isinstance(element.data, Unset):
                table.append((key, "Not set"))
            else:
                if isinstance(element.decoded, Unset):
                    table.append((key, element.data))
                else:
                    self._kv_display(table, key, element.decoded.additional_properties)

        print(tabulate(table))


class Coap(CloudSubCommand):
    @classmethod
    def add_parser(cls, parser):
        parser_coap = parser.add_parser("coap", help="CoAP file server")
        parser_coap.set_defaults(command_class=cls)

        tool_parser = parser_coap.add_subparsers(title="commands", metavar="<command>", required=True)

        list_parser = tool_parser.add_parser("list")
        list_parser.set_defaults(command_fn=cls.list)

    def run(self):
        with self.client() as client:
            self.args.command_fn(self, client)

    def list(self, client: Client):
        files = get_coap_files.sync(client=client)

        if isinstance(files, COAPFilesList):
            sorted_list: list[str] = sorted(files.filenames)
            print("CoAP Files:")
            print("\t" + "\n\t".join(sorted_list))
        else:
            print(f"Failed to retrieve file list {files}")


class SubCommand(InfuseCommand):
    NAME = "cloud"
    HELP = "Infuse-IoT cloud interaction"
    DESCRIPTION = "Infuse-IoT cloud interaction"

    @classmethod
    def add_parser(cls, parser):
        subparser = parser.add_subparsers(title="commands", metavar="<command>", required=True)

        Organisations.add_parser(subparser)
        Boards.add_parser(subparser)
        Device.add_parser(subparser)
        Coap.add_parser(subparser)

    def __init__(self, args):
        self.tool = args.command_class(args)

    def run(self):
        self.tool.run()



#!/usr/bin/env python3

import infuse_iot.generated.tdf_definitions as _tdf_base
from infuse_iot.util.internal import extension_load as _loader

if _extension_module := _loader("tdf_definitions"):
    id_type_mapping = {
        **_tdf_base.id_type_mapping,
        **_extension_module.id_type_mapping,
    }

    class readings(_tdf_base.readings, _extension_module.readings):  # type: ignore
        pass

    class structs(_tdf_base.structs, _extension_module.structs):  # type: ignore
        pass
else:
    id_type_mapping = _tdf_base.id_type_mapping
    readings = _tdf_base.readings  # type: ignore
    structs = _tdf_base.structs  # type: ignore


__all__ = [
    "id_type_mapping",
    "readings",
    "structs",
]



#!/usr/bin/env python3

import threading
from collections.abc import Callable


class SignaledThread(threading.Thread):
    """Thread that can be signaled to terminate"""

    def __init__(self, fn: Callable[[], None], sig: threading.Event | None = None):
        self._fn = fn
        self._sig = sig if sig is not None else threading.Event()
        super().__init__(target=self.run_loop)

    def stop(self):
        """Signal thread to terminate"""
        self._sig.set()

    def run_loop(self):
        """Run the thread function in a loop"""
        try:
            while not self._sig.is_set():
                self._fn()
        except Exception as e:
            import traceback

            # Print the exception traceback
            traceback.print_exception(e)
            # Set the termination signal
            self._sig.set()



#!/usr/bin/env python3

import base64
import ctypes
import enum
import random
import time
from typing import Any

from typing_extensions import Self

from infuse_iot.common import InfuseID, InfuseType
from infuse_iot.database import DeviceDatabase, NoKeyError
from infuse_iot.epacket.common import Serializable
from infuse_iot.epacket.interface import ID as Interface
from infuse_iot.epacket.interface import Address
from infuse_iot.time import InfuseTime
from infuse_iot.util.crypto import chachapoly_decrypt, chachapoly_encrypt


class Auth(enum.IntEnum):
    """Authorisation options"""

    DEVICE = 0
    NETWORK = 1


class Flags(enum.IntEnum):
    ENCR_DEVICE = 0x8000
    ENCR_NETWORK = 0x0000


class HopOutput(Serializable):
    def __init__(self, infuse_id: int, interface: Interface, auth: Auth):
        self.infuse_id = infuse_id
        self.interface = interface
        self.auth = auth

    def to_json(self) -> dict:
        return {
            "id": self.infuse_id,
            "interface": self.interface.value,
            "auth": self.auth.value,
        }

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(
            infuse_id=values["id"],
            interface=Interface(values["interface"]),
            auth=Auth(values["auth"]),
        )

    @classmethod
    def serial(cls, auth=Auth.DEVICE) -> Self:
        """Local serial hop"""
        return cls(0, Interface.SERIAL, auth)


class HopReceived(Serializable):
    def __init__(
        self,
        infuse_id: int,
        interface: Interface,
        interface_address: Address,
        auth: Auth,
        key_identifier: int,
        gps_time: int,
        sequence: int,
        rssi: int,
    ):
        self.infuse_id = infuse_id
        self.interface = interface
        self.interface_address = interface_address
        self.auth = auth
        self.key_identifier = key_identifier
        self.gps_time = gps_time
        self.sequence = sequence
        self.rssi = rssi

    def to_json(self) -> dict:
        return {
            "id": self.infuse_id,
            "interface": self.interface.value,
            "interface_addr": self.interface_address.to_json(),
            "auth": self.auth.value,
            "key_id": self.key_identifier,
            "time": self.gps_time,
            "seq": self.sequence,
            "rssi": self.rssi,
        }

    @classmethod
    def from_json(cls, values: dict) -> Self:
        interface = Interface(values["interface"])
        return cls(
            infuse_id=values["id"],
            interface=interface,
            interface_address=Address.from_json(values["interface_addr"]),
            auth=Auth(values["auth"]),
            key_identifier=values["key_id"],
            gps_time=values["time"],
            sequence=values["seq"],
            rssi=values["rssi"],
        )


class PacketReceived(Serializable):
    """ePacket received by a gateway"""

    def __init__(self, route: list[HopReceived], ptype: InfuseType, payload: bytes):
        # [Original Transmission, hop, hop, serial]
        self.route = route
        self.ptype = ptype
        self.payload = payload

    def to_json(self) -> dict:
        """Convert class to json dictionary"""
        return {
            "route": [x.to_json() for x in self.route],
            "type": self.ptype.value,
            "payload": base64.b64encode(self.payload).decode("utf-8"),
        }

    @classmethod
    def from_json(cls, values: dict) -> Self:
        """Reconstruct class from json dictionary"""
        return cls(
            route=[HopReceived.from_json(x) for x in values["route"]],
            ptype=InfuseType(values["type"]),
            payload=base64.b64decode(values["payload"].encode("utf-8")),
        )

    @classmethod
    def from_serial(cls, database: DeviceDatabase, serial_frame: bytes) -> list[Self]:
        header, decrypted = CtypeSerialFrame.decrypt(database, serial_frame)

        # Packet from local gateway
        if header.type != InfuseType.RECEIVED_EPACKET:
            return [cls([header.hop_received()], header.type, decrypted)]

        # Extract packets contained in payload
        packets = []
        buffer = bytearray(decrypted)
        while len(buffer) > 0:
            common_header = CtypePacketReceived.CommonHeader.from_buffer_copy(buffer)
            packet_bytes = buffer[: common_header.len]
            del buffer[: common_header.len]
            del packet_bytes[: ctypes.sizeof(common_header)]

            # Only Bluetooth advertising supported for now
            decode_mapping: dict[Interface, Any] = {
                Interface.BT_ADV: CtypeBtAdvFrame,
                Interface.BT_PERIPHERAL: CtypeBtGattFrame,
                Interface.BT_CENTRAL: CtypeBtGattFrame,
            }
            if common_header.interface not in decode_mapping:
                raise NotImplementedError
            frame_type = decode_mapping[common_header.interface]

            # Extract interface address (Only Bluetooth supported)
            addr = Address.from_bytes(common_header.interface, packet_bytes)
            del packet_bytes[: addr.len()]

            # Decrypting packet
            if common_header.encrypted:
                try:
                    f_header, f_decrypted = frame_type.decrypt(database, addr.val, packet_bytes)
                except NoKeyError:
                    continue

                bt_hop = HopReceived(
                    f_header.device_id,
                    common_header.interface,
                    addr,
                    (Auth.DEVICE if f_header.flags & Flags.ENCR_DEVICE else Auth.NETWORK),
                    f_header.key_metadata,
                    f_header.gps_time,
                    f_header.sequence,
                    common_header.rssi,
                )
                packet = cls(
                    [bt_hop, header.hop_received()],
                    f_header.type,
                    bytes(f_decrypted),
                )
            else:
                # Extract payload metadata
                decr_header = CtypePacketReceived.DecryptedHeader.from_buffer_copy(packet_bytes)
                del packet_bytes[: ctypes.sizeof(decr_header)]

                # Notify database of BT Addr -> Infuse ID mapping
                database.observe_device(decr_header.device_id, bt_addr=addr.val)

                bt_hop = HopReceived(
                    decr_header.device_id,
                    common_header.interface,
                    addr,
                    (Auth.DEVICE if decr_header.flags & Flags.ENCR_DEVICE else Auth.NETWORK),
                    decr_header.key_id,
                    decr_header.gps_time,
                    decr_header.sequence,
                    common_header.rssi,
                )
                packet = cls(
                    [bt_hop, header.hop_received()],
                    decr_header.type,
                    bytes(packet_bytes),
                )
            packets.append(packet)

        return packets


class PacketOutputRouted(Serializable):
    """ePacket to be transmitted by gateway with complete route"""

    def __init__(self, route: list[HopOutput], ptype: InfuseType, payload: bytes):
        # [Serial, hop, hop, final_hop]
        self.route = route
        self.ptype = ptype
        self.payload = payload

    def to_serial(self, database: DeviceDatabase) -> bytes:
        """Encode and encrypt packet for serial transmission"""
        gps_time = InfuseTime.gps_seconds_from_unix(int(time.time()))

        if len(self.route) == 2:
            # Two hops only supports Bluetooth central for now
            final = self.route[1]
            bt_addr = database.devices[final.infuse_id].bt_addr
            assert final.interface == Interface.BT_CENTRAL
            assert bt_addr is not None

            # Forwarded payload
            forward_payload = CtypeBtGattFrame.encrypt(database, final.infuse_id, self.ptype, Auth.DEVICE, self.payload)

            # Forwarding header
            forward_hdr = CtypeForwardHeaderBtGatt(
                ctypes.sizeof(CtypeForwardHeaderBtGatt) + len(forward_payload),
                Interface.BT_CENTRAL.value,
                bt_addr.to_ctype(),
            )

            ptype = InfuseType.EPACKET_FORWARD
            payload = bytes(forward_hdr) + forward_payload
        elif len(self.route) == 1:
            ptype = self.ptype
            payload = self.payload
        else:
            raise NotImplementedError(">2 hops currently not supported")

        serial = self.route[0]

        if serial.auth == Auth.NETWORK:
            flags = Flags.ENCR_NETWORK
            key_metadata = database.devices[serial.infuse_id].network_id
            key = database.serial_network_key(serial.infuse_id, gps_time)
        else:
            flags = Flags.ENCR_DEVICE
            key_metadata = database.devices[serial.infuse_id].device_id
            key = database.serial_device_key(serial.infuse_id, gps_time)

        # Validation
        assert key_metadata is not None

        # Create header
        header = CtypeSerialFrame(
            version=0,
            _type=ptype,
            flags=flags,
            gps_time=gps_time,
            sequence=0,
            entropy=random.randint(0, 65535),
        )
        header.key_metadata = key_metadata
        if serial.infuse_id == InfuseID.GATEWAY:
            assert database.gateway is not None
            header.device_id = database.gateway
        else:
            header.device_id = serial.infuse_id

        # Encrypt and return payload
        header_bytes = bytes(header)
        ciphertext = chachapoly_encrypt(key, header_bytes[:11], header_bytes[11:], payload)
        return header_bytes + ciphertext

    def to_json(self) -> dict:
        return {
            "route": [x.to_json() for x in self.route],
            "type": self.ptype.value,
            "payload": base64.b64encode(self.payload).decode("utf-8"),
        }

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(
            route=[HopOutput.from_json(x) for x in values["route"]],
            ptype=InfuseType(values["type"]),
            payload=base64.b64decode(values["payload"].encode("utf-8")),
        )


class PacketOutput(PacketOutputRouted):
    """ePacket to be transmitted by gateway"""

    def __init__(self, infuse_id: int, auth: Auth, ptype: InfuseType, payload: bytes):
        self.infuse_id = infuse_id
        self.auth = auth
        self.ptype = ptype
        self.payload = payload

    def to_json(self) -> dict:
        return {
            "infuse_id": self.infuse_id,
            "auth": self.auth,
            "type": self.ptype.value,
            "payload": base64.b64encode(self.payload).decode("utf-8"),
        }

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(
            infuse_id=values["infuse_id"],
            auth=Auth(values["auth"]),
            ptype=InfuseType(values["type"]),
            payload=base64.b64decode(values["payload"].encode("utf-8")),
        )


class CtypeForwardHeaderBtGatt(ctypes.LittleEndianStructure):
    _fields_ = [
        ("total_length", ctypes.c_uint16),
        ("interface", ctypes.c_uint8),
        ("address", Address.BluetoothLeAddr.CtypesFormat),
    ]
    _pack_ = 1


class CtypeV0Frame(ctypes.LittleEndianStructure):
    _fields_ = []
    _pack_ = 1

    @property
    def type(self) -> InfuseType:
        return InfuseType(self._type)

    @property
    def key_metadata(self) -> int:
        return int.from_bytes(self._key_metadata, "little")

    @key_metadata.setter
    def key_metadata(self, value):
        self._key_metadata[:] = value.to_bytes(3, "little")

    @property
    def device_id(self) -> int:
        return (self._device_id_upper << 32) | self._device_id_lower

    @device_id.setter
    def device_id(self, value):
        self._device_id_upper = value >> 32
        self._device_id_lower = value & 0xFFFFFFFF

    @classmethod
    def parse(cls, frame: bytes) -> tuple[Self, int]:
        """Parse frame into header and payload length"""
        return (
            cls.from_buffer_copy(frame),
            len(frame) - ctypes.sizeof(CtypeV0VersionedFrame) - 16,
        )


class CtypeV0VersionedFrame(CtypeV0Frame):
    _fields_ = [
        ("version", ctypes.c_uint8),
        ("_type", ctypes.c_uint8),
        ("flags", ctypes.c_uint16),
        ("_key_metadata", ctypes.c_uint8 * 3),
        ("_device_id_upper", ctypes.c_uint32),
        ("_device_id_lower", ctypes.c_uint32),
        ("gps_time", ctypes.c_uint32),
        ("sequence", ctypes.c_uint16),
        ("entropy", ctypes.c_uint16),
    ]
    _pack_ = 1


class CtypeV0UnversionedFrame(CtypeV0Frame):
    _fields_ = [
        ("_type", ctypes.c_uint8),
        ("flags", ctypes.c_uint16),
        ("_key_metadata", ctypes.c_uint8 * 3),
        ("_device_id_upper", ctypes.c_uint32),
        ("_device_id_lower", ctypes.c_uint32),
        ("gps_time", ctypes.c_uint32),
        ("sequence", ctypes.c_uint16),
        ("entropy", ctypes.c_uint16),
    ]
    _pack_ = 1


class CtypeSerialFrame(CtypeV0VersionedFrame):
    """Serial packet header"""

    def hop_received(self) -> HopReceived:
        auth = Auth.DEVICE if self.flags & Flags.ENCR_DEVICE else Auth.NETWORK
        return HopReceived(
            self.device_id,
            Interface.SERIAL,
            Address(Address.SerialAddr()),
            auth,
            self.key_metadata,
            self.gps_time,
            self.sequence,
            0,
        )

    @classmethod
    def decrypt(cls, database: DeviceDatabase, frame: bytes):
        header = cls.from_buffer_copy(frame)
        if header.flags & Flags.ENCR_DEVICE:
            database.observe_device(header.device_id, device_id=header.key_metadata)
            key = database.serial_device_key(header.device_id, header.gps_time)
        else:
            database.observe_device(header.device_id, network_id=header.key_metadata)
            key = database.serial_network_key(header.device_id, header.gps_time)

        decrypted = chachapoly_decrypt(key, frame[:11], frame[11:23], frame[23:])
        return header, decrypted


class CtypeBtAdvFrame(CtypeV0VersionedFrame):
    """Bluetooth Advertising packet header"""

    @classmethod
    def decrypt(cls, database: DeviceDatabase, bt_addr: Address.BluetoothLeAddr, frame: bytes):
        header = cls.from_buffer_copy(frame)
        if header.flags & Flags.ENCR_DEVICE:
            raise NotImplementedError
        else:
            database.observe_device(header.device_id, network_id=header.key_metadata, bt_addr=bt_addr)
            key = database.bt_adv_network_key(header.device_id, header.gps_time)

        decrypted = chachapoly_decrypt(key, frame[:11], frame[11:23], frame[23:])
        return header, decrypted


class CtypeBtGattFrame(CtypeV0VersionedFrame):
    """Bluetooth GATT packet header"""

    @classmethod
    def encrypt(
        cls,
        database: DeviceDatabase,
        infuse_id: int,
        ptype: InfuseType,
        auth: Auth,
        payload: bytes,
    ) -> bytes:
        dev_state = database.devices[infuse_id]
        gps_time = InfuseTime.gps_seconds_from_unix(int(time.time()))
        flags = 0

        if auth == Auth.DEVICE:
            key_meta = dev_state.device_id
            key = database.bt_gatt_device_key(infuse_id, gps_time)
            flags |= Flags.ENCR_DEVICE
        else:
            key_meta = dev_state.network_id
            key = database.bt_gatt_network_key(infuse_id, gps_time)

        # Validate
        assert key_meta is not None

        # Construct GATT header
        header = cls(
            _type=ptype,
            flags=flags,
            gps_time=gps_time,
            sequence=dev_state.gatt_sequence_num(),
            entropy=random.randint(0, 65535),
        )
        header.device_id = infuse_id
        header.key_metadata = key_meta

        # Encrypt and return payload
        header_bytes = bytes(header)
        ciphertext = chachapoly_encrypt(key, header_bytes[:11], header_bytes[11:], payload)
        return header_bytes + ciphertext

    @classmethod
    def decrypt(cls, database: DeviceDatabase, bt_addr: Address.BluetoothLeAddr | None, frame: bytes):
        header = cls.from_buffer_copy(frame)
        if header.flags & Flags.ENCR_DEVICE:
            database.observe_device(header.device_id, device_id=header.key_metadata, bt_addr=bt_addr)
            key = database.bt_gatt_device_key(header.device_id, header.gps_time)
        else:
            database.observe_device(header.device_id, network_id=header.key_metadata, bt_addr=bt_addr)
            key = database.bt_gatt_network_key(header.device_id, header.gps_time)

        decrypted = chachapoly_decrypt(key, frame[:11], frame[11:23], frame[23:])
        return header, decrypted


class CtypeUdpFrame(CtypeV0UnversionedFrame):
    @classmethod
    def decrypt(cls, database: DeviceDatabase, frame: bytes):
        header = cls.from_buffer_copy(frame)
        if header.flags & Flags.ENCR_DEVICE:
            database.observe_device(header.device_id, device_id=header.key_metadata)
            key = database.udp_device_key(header.device_id, header.gps_time)
        else:
            database.observe_device(header.device_id, network_id=header.key_metadata)
            key = database.udp_network_key(header.device_id, header.gps_time)

        decrypted = chachapoly_decrypt(key, frame[:10], frame[10:22], frame[22:])
        return header, decrypted


class CtypePacketReceived:
    class CommonHeader(ctypes.Structure):
        _fields_ = [
            ("_len_encr", ctypes.c_uint16),
            ("_rssi", ctypes.c_uint8),
            ("_if", ctypes.c_uint8),
        ]
        _pack_ = 1

        @property
        def len(self):
            return self._len_encr & 0x7FFF

        @property
        def encrypted(self):
            return (self._len_encr & 0x8000) != 0

        @property
        def interface(self):
            return Interface(self._if)

        @property
        def rssi(self):
            return 0 - self._rssi

    class DecryptedHeader(ctypes.Structure):
        _fields_ = [
            ("device_id", ctypes.c_uint64),
            ("gps_time", ctypes.c_uint32),
            ("_type", ctypes.c_uint8),
            ("flags", ctypes.c_uint16),
            ("sequence", ctypes.c_uint16),
            ("_key_id", 3 * ctypes.c_uint8),
        ]
        _pack_ = 1

        @property
        def type(self):
            return InfuseType(self._type)

        @property
        def key_id(self):
            return int.from_bytes(self._key_id, "little")



#!/usr/bin/env python3

import datetime
import enum


class InfuseTimeSource:
    class base(enum.IntEnum):
        NONE = 0
        GNSS = 1
        NTP = 2
        RPC = 3
        INVALID = 4

    def __init__(self, value: int):
        self.recovered: bool = value & 0x80 != 0
        self.source = self.base(value & 0x7F)

    def __str__(self) -> str:
        postfix = " (recovered after reboot)" if self.recovered else ""
        base = self.source.name if self.source.name is not None else "Unknown"
        return base + postfix


class InfuseTime:
    GPS_UNIX_OFFSET = 315964800
    UNIX_LEAP_SECONDS = 18

    @classmethod
    def unix_time_from_gps_seconds(cls, gps_seconds: int) -> int:
        return gps_seconds + cls.GPS_UNIX_OFFSET - cls.UNIX_LEAP_SECONDS

    @classmethod
    def unix_time_from_epoch(cls, epoch_time: int) -> float:
        whole = epoch_time // 65536
        partial = epoch_time % 65536
        return cls.unix_time_from_gps_seconds(whole) + (partial / 65536)

    @classmethod
    def gps_seconds_from_unix(cls, unix_seconds: int) -> int:
        return unix_seconds - cls.GPS_UNIX_OFFSET + cls.UNIX_LEAP_SECONDS

    @classmethod
    def epoch_time_from_unix(cls, unix_time: float) -> int:
        whole = int(unix_time)
        frac = unix_time - whole
        return (cls.gps_seconds_from_unix(whole) * 65536) + int(frac * 65536)

    @classmethod
    def utc_time_string(cls, unix_time: float) -> str:
        # Trim timezone prefix and microseconds
        return str(datetime.datetime.fromtimestamp(unix_time, datetime.timezone.utc))[:-9]

    @classmethod
    def utc_time_string_log(cls, unix_time: float) -> str:
        obj = datetime.datetime.fromtimestamp(unix_time, datetime.timezone.utc)
        return obj.strftime("%Y-%m-%dT%H:%M:%S.%f")



#!/usr/bin/env python3


import datetime
import threading

import colorama

try:
    from simple_term_menu import TerminalMenu
except NotImplementedError:
    pass

_lock = threading.Lock()


class Console:
    """Common terminal logging functions"""

    @staticmethod
    def init():
        """Initialise console logging"""
        colorama.init(autoreset=True)

    @staticmethod
    def log_error(message):
        """Log error message to terminal"""
        Console.log(datetime.datetime.now(), colorama.Fore.RED, message)

    @staticmethod
    def log_info(message):
        """Log info message to terminal"""
        Console.log(datetime.datetime.now(), colorama.Fore.MAGENTA, message)

    @staticmethod
    def log_text(message):
        """Log text to terminal"""
        Console.log(datetime.datetime.now(), colorama.Fore.RESET, message)

    @staticmethod
    def log_tx(data_type, length, prefix=""):
        """Log transmitted packet to terminal"""
        Console.log(
            datetime.datetime.now(),
            colorama.Fore.BLUE,
            f"{prefix}TX {data_type.name} {length} bytes",
        )

    @staticmethod
    def log_rx(data_type, length, prefix=""):
        """Log received packet to terminal"""
        Console.log(
            datetime.datetime.now(),
            colorama.Fore.GREEN,
            f"{prefix}RX {data_type.name} {length} bytes",
        )

    @staticmethod
    def log(timestamp: datetime.datetime, colour, string: str):
        """Log colourised string to terminal"""
        ts = timestamp.strftime("%H:%M:%S.%f")[:-3]
        with _lock:
            print(f"[{ts}]{colour} {string}{colorama.Fore.RESET}")


def choose_one(title: str, options: list[str]) -> tuple[int, str]:
    """Select a single option from a list"""

    if TerminalMenu:
        # Linux & MacOS
        terminal_menu = TerminalMenu(options, title=title)
        idx = terminal_menu.show()
        if idx is None:
            raise IndexError("No option chosen")
        return idx, options[idx]
    else:
        # Windows
        print(title)
        for idx, option in enumerate(options):
            print(f" {idx:2d}: {option}")
        idx = None
        while idx is None:
            try:
                idx = int(input(f"Enter index between 0 and {len(options) - 1}:"))
                if not (0 <= idx < len(options)):
                    idx = None
            except ValueError:
                pass

        return idx, options[idx]



#!/usr/bin/env python3

import ctypes

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.util.ctypes import VLACompatLittleEndianStruct
from infuse_iot.zephyr.errno import errno


class kv_reflect_crcs(InfuseRpcCommand, defs.kv_reflect_crcs):
    HELP = "Read KV store reflection crc values"
    DESCRIPTION = "Read KV store reflection crc values"

    class response(VLACompatLittleEndianStruct):
        _fields_ = [
            ("num", ctypes.c_uint16),
            ("remaining", ctypes.c_uint16),
        ]
        vla_field = ("crcs", 0 * defs.rpc_struct_kv_store_crc)
        _pack_ = 1

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--offset", type=int, default=0, help="Offset to start CRC read at")

    def __init__(self, args):
        self.offset = args.offset

    def request_struct(self):
        return self.request(self.offset)

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Invalid query ({errno.strerror(-return_code)})")
            return

        print(f"Slot CRCs ({response.remaining} remaining):")
        for slot in response.crcs:
            print(f"\t{slot.id:5d}: 0x{slot.crc:08x}")



import datetime
from http import HTTPStatus
from typing import Any
from uuid import UUID

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.downlink_message_status import DownlinkMessageStatus
from ...models.error import Error
from ...models.rpc_message import RpcMessage
from ...types import UNSET, Response, Unset


def _get_kwargs(
    *,
    organisation_id: Unset | UUID = UNSET,
    device_id: Unset | str = UNSET,
    status: Unset | DownlinkMessageStatus = UNSET,
    start_time: Unset | datetime.datetime = UNSET,
    end_time: Unset | datetime.datetime = UNSET,
    limit: Unset | int = 10,
    rpc_command_id: Unset | int = UNSET,
    show_expired: Unset | bool = True,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    json_organisation_id: Unset | str = UNSET
    if not isinstance(organisation_id, Unset):
        json_organisation_id = str(organisation_id)
    params["organisationId"] = json_organisation_id

    params["deviceId"] = device_id

    json_status: Unset | str = UNSET
    if not isinstance(status, Unset):
        json_status = status.value

    params["status"] = json_status

    json_start_time: Unset | str = UNSET
    if not isinstance(start_time, Unset):
        json_start_time = start_time.isoformat()
    params["startTime"] = json_start_time

    json_end_time: Unset | str = UNSET
    if not isinstance(end_time, Unset):
        json_end_time = end_time.isoformat()
    params["endTime"] = json_end_time

    params["limit"] = limit

    params["rpcCommandId"] = rpc_command_id

    params["showExpired"] = show_expired

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": "/rpc",
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: AuthenticatedClient | Client, response: httpx.Response
) -> Error | list["RpcMessage"] | None:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = RpcMessage.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if response.status_code == 500:
        response_500 = Error.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: AuthenticatedClient | Client, response: httpx.Response
) -> Response[Error | list["RpcMessage"]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient | Client,
    organisation_id: Unset | UUID = UNSET,
    device_id: Unset | str = UNSET,
    status: Unset | DownlinkMessageStatus = UNSET,
    start_time: Unset | datetime.datetime = UNSET,
    end_time: Unset | datetime.datetime = UNSET,
    limit: Unset | int = 10,
    rpc_command_id: Unset | int = UNSET,
    show_expired: Unset | bool = True,
) -> Response[Error | list["RpcMessage"]]:
    """Get RPC messages

    Args:
        organisation_id (Union[Unset, UUID]): ID of organisation
        device_id (Union[Unset, str]): 8 byte DeviceID as a hex string (if not provided will be
            auto-generated) Example: d291d4d66bf0a955.
        status (Union[Unset, DownlinkMessageStatus]): Status of downlink message
        start_time (Union[Unset, datetime.datetime]): The start time of the query (only return
            items on or after this time)
        end_time (Union[Unset, datetime.datetime]): The end time of the query (only return items
            on or before this time)
        limit (Union[Unset, int]): Maximum number of items to return Default: 10.
        rpc_command_id (Union[Unset, int]): ID of RPC command
        show_expired (Union[Unset, bool]): Whether to show expired RPC messages Default: True.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, list['RpcMessage']]]
    """

    kwargs = _get_kwargs(
        organisation_id=organisation_id,
        device_id=device_id,
        status=status,
        start_time=start_time,
        end_time=end_time,
        limit=limit,
        rpc_command_id=rpc_command_id,
        show_expired=show_expired,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient | Client,
    organisation_id: Unset | UUID = UNSET,
    device_id: Unset | str = UNSET,
    status: Unset | DownlinkMessageStatus = UNSET,
    start_time: Unset | datetime.datetime = UNSET,
    end_time: Unset | datetime.datetime = UNSET,
    limit: Unset | int = 10,
    rpc_command_id: Unset | int = UNSET,
    show_expired: Unset | bool = True,
) -> Error | list["RpcMessage"] | None:
    """Get RPC messages

    Args:
        organisation_id (Union[Unset, UUID]): ID of organisation
        device_id (Union[Unset, str]): 8 byte DeviceID as a hex string (if not provided will be
            auto-generated) Example: d291d4d66bf0a955.
        status (Union[Unset, DownlinkMessageStatus]): Status of downlink message
        start_time (Union[Unset, datetime.datetime]): The start time of the query (only return
            items on or after this time)
        end_time (Union[Unset, datetime.datetime]): The end time of the query (only return items
            on or before this time)
        limit (Union[Unset, int]): Maximum number of items to return Default: 10.
        rpc_command_id (Union[Unset, int]): ID of RPC command
        show_expired (Union[Unset, bool]): Whether to show expired RPC messages Default: True.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, list['RpcMessage']]
    """

    return sync_detailed(
        client=client,
        organisation_id=organisation_id,
        device_id=device_id,
        status=status,
        start_time=start_time,
        end_time=end_time,
        limit=limit,
        rpc_command_id=rpc_command_id,
        show_expired=show_expired,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient | Client,
    organisation_id: Unset | UUID = UNSET,
    device_id: Unset | str = UNSET,
    status: Unset | DownlinkMessageStatus = UNSET,
    start_time: Unset | datetime.datetime = UNSET,
    end_time: Unset | datetime.datetime = UNSET,
    limit: Unset | int = 10,
    rpc_command_id: Unset | int = UNSET,
    show_expired: Unset | bool = True,
) -> Response[Error | list["RpcMessage"]]:
    """Get RPC messages

    Args:
        organisation_id (Union[Unset, UUID]): ID of organisation
        device_id (Union[Unset, str]): 8 byte DeviceID as a hex string (if not provided will be
            auto-generated) Example: d291d4d66bf0a955.
        status (Union[Unset, DownlinkMessageStatus]): Status of downlink message
        start_time (Union[Unset, datetime.datetime]): The start time of the query (only return
            items on or after this time)
        end_time (Union[Unset, datetime.datetime]): The end time of the query (only return items
            on or before this time)
        limit (Union[Unset, int]): Maximum number of items to return Default: 10.
        rpc_command_id (Union[Unset, int]): ID of RPC command
        show_expired (Union[Unset, bool]): Whether to show expired RPC messages Default: True.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, list['RpcMessage']]]
    """

    kwargs = _get_kwargs(
        organisation_id=organisation_id,
        device_id=device_id,
        status=status,
        start_time=start_time,
        end_time=end_time,
        limit=limit,
        rpc_command_id=rpc_command_id,
        show_expired=show_expired,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient | Client,
    organisation_id: Unset | UUID = UNSET,
    device_id: Unset | str = UNSET,
    status: Unset | DownlinkMessageStatus = UNSET,
    start_time: Unset | datetime.datetime = UNSET,
    end_time: Unset | datetime.datetime = UNSET,
    limit: Unset | int = 10,
    rpc_command_id: Unset | int = UNSET,
    show_expired: Unset | bool = True,
) -> Error | list["RpcMessage"] | None:
    """Get RPC messages

    Args:
        organisation_id (Union[Unset, UUID]): ID of organisation
        device_id (Union[Unset, str]): 8 byte DeviceID as a hex string (if not provided will be
            auto-generated) Example: d291d4d66bf0a955.
        status (Union[Unset, DownlinkMessageStatus]): Status of downlink message
        start_time (Union[Unset, datetime.datetime]): The start time of the query (only return
            items on or after this time)
        end_time (Union[Unset, datetime.datetime]): The end time of the query (only return items
            on or before this time)
        limit (Union[Unset, int]): Maximum number of items to return Default: 10.
        rpc_command_id (Union[Unset, int]): ID of RPC command
        show_expired (Union[Unset, bool]): Whether to show expired RPC messages Default: True.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, list['RpcMessage']]
    """

    return (
        await asyncio_detailed(
            client=client,
            organisation_id=organisation_id,
            device_id=device_id,
            status=status,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            rpc_command_id=rpc_command_id,
            show_expired=show_expired,
        )
    ).parsed



#!/usr/bin/env python3

import ipaddress

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr import lte as z_lte
from infuse_iot.zephyr import net_if as z_nif
from infuse_iot.zephyr.errno import errno


class lte_state(InfuseRpcCommand, defs.lte_state):
    @classmethod
    def add_parser(cls, parser):
        return

    def __init__(self, args):
        self.args = args

    def request_struct(self):
        return self.request()

    def request_json(self):
        return {}

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query current time ({errno.strerror(-return_code)})")
            return

        common = response.common
        lte = response.lte

        # Address formatting
        ipv4 = ipaddress.IPv4Address(bytes(common.ipv4))
        ipv6 = ipaddress.IPv6Address(bytes(common.ipv6))

        reg_class = z_lte.RegistrationState
        lte_state = reg_class(lte.registration_state)

        print("Interface State:")
        print(f"\t          State: {z_nif.OperationalState(common.state).name}")
        print(f"\t       IF Flags: {z_nif.InterfaceFlags(common.if_flags)}")
        print(f"\t       L2 Flags: {z_nif.L2Flags(common.l2_flags)}")
        print(f"\t            MTU: {common.mtu}")
        print(f"\t           IPv4: {ipv4}")
        print(f"\t           IPv6: {ipv6}")
        print("LTE State:")
        print(f"\t      Reg State: {lte_state}")

        valid = (
            lte_state == reg_class.REGISTERED_HOME
            or lte_state == reg_class.REGISTERED_ROAMING
            or lte_state == reg_class.SEARCHING
        )
        access_tech = z_lte.AccessTechnology(lte.access_technology)
        if valid:
            if lte.earfcn != 0:
                freq_dl, freq_ul = z_lte.LteBands.earfcn_to_freq(lte.earfcn)
                freq_string = f" (UL: {int(freq_ul)}MHz, DL: {int(freq_dl)}MHz)"
            else:
                freq_string = ""
            country = z_lte.MobileCountryCodes.name_from_mcc(lte.mcc)
            active_str = f"{lte.psm_active_time} s" if lte.psm_active_time != 65535 else "N/A"
            edrx_interval_str = f"{lte.edrx_interval} s" if lte.edrx_interval != -1.0 else "N/A"
            edrx_window_str = f"{lte.edrx_paging_window} s" if lte.edrx_paging_window != -1.0 else "N/A"
            print(f"\t    Access Tech: {access_tech}")
            print(f"\t   Country Code: {lte.mcc} ({country})")
            print(f"\t   Network Code: {lte.mnc}")
            print(f"\t        Cell ID: {lte.cell_id}")
            print(f"\t  Tracking Area: {lte.tac}")
            print(f"\t            TAU: {lte.tau} s")
            print(f"\t         EARFCN: {lte.earfcn}{freq_string}")
            print(f"\t           Band: {lte.band}")
            print(f"\tPSM Active Time: {active_str}")
            print(f"\t  eDRX Interval: {edrx_interval_str}")
            print(f"\t    eDRX Window: {edrx_window_str}")
            print(f"\t           RSRP: {lte.rsrp} dBm")
            print(f"\t           RSRQ: {lte.rsrq} dB")



#!/usr/bin/env python3

"""Automatically OTA upgrade observed devices"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import argparse
import binascii
import sys
import time

from rich.live import Live
from rich.progress import (
    DownloadColumn,
    Progress,
    TransferSpeedColumn,
)
from rich.status import Status
from rich.table import Table

from infuse_iot.commands import InfuseCommand
from infuse_iot.common import InfuseID
from infuse_iot.definitions.rpc import bt_file_copy_basic, file_write_basic, rpc_enum_file_action
from infuse_iot.epacket.packet import Auth, HopReceived
from infuse_iot.rpc_client import RpcClient
from infuse_iot.socket_comms import (
    GatewayRequestConnectionRequest,
    LocalClient,
    default_multicast_address,
)
from infuse_iot.util.argparse import ValidFile, ValidRelease
from infuse_iot.zephyr.errno import errno


class SubCommand(InfuseCommand):
    NAME = "ota_upgrade"
    HELP = "Automatically OTA upgrade observed devices"
    DESCRIPTION = "Automatically OTA upgrade observed devices"

    def __init__(self, args):
        self._client = LocalClient(default_multicast_address(), 1.0)
        self._conn_timeout = args.conn_timeout
        self._min_rssi: int | None = args.rssi
        self._explicit_ids: list[int] = []
        if args.release:
            self._release: ValidRelease = args.release
            self._single_diff = None
        elif args.single:
            # Find the associated release
            diff_folder = args.single.parent
            if diff_folder.name != "diffs":
                # Try the next level up
                diff_folder = diff_folder.parent
            if diff_folder.name != "diffs":
                raise argparse.ArgumentTypeError(f"{args.single} is not in a diff (sub)folder")
            release_folder = diff_folder.parent
            self._release = ValidRelease(str(release_folder))
            self._single_diff = args.single
        else:
            raise NotImplementedError("Unknow upgrade type")
        self._app_name = self._release.metadata["application"]["primary"]
        self._app_id = self._release.metadata["application"]["id"]
        self._new_ver = self._release.metadata["application"]["version"]
        self._handled: list[int] = []
        self._pending: dict[int, float] = {}
        self._missing_diffs: set[str] = set()
        self._already = 0
        self._updated = 0
        self._no_diff = 0
        self._failed = 0
        self.patch_file = b""
        self.state = "Scanning"
        self.progress = Progress(
            *Progress.get_default_columns(),
            DownloadColumn(),
            TransferSpeedColumn(),
        )
        self.task = None
        if args.log is None:
            self._log = None
        else:
            self._log = open(args.log, "+a", encoding="utf-8")  # noqa: SIM115

        if args.id is not None:
            self._explicit_ids.append(args.id)
        elif args.list is not None:
            with args.list.open("r") as f:
                for line in f.readlines():
                    self._explicit_ids.append(int(line.strip(), 0))

    @classmethod
    def add_parser(cls, parser):
        upgrade_type = parser.add_mutually_exclusive_group(required=True)
        upgrade_type.add_argument("--release", "-r", type=ValidRelease, help="Application release to upgrade to")
        upgrade_type.add_argument("--single", type=ValidFile, help="Single diff")
        parser.add_argument("--rssi", type=int, help="Minimum RSSI to attempt upgrade process")
        parser.add_argument("--log", type=str, help="File to write upgrade results to")
        parser.add_argument(
            "--conn-timeout", type=int, default=10000, help="Timeout to wait for a connection to the device (ms)"
        )
        explicit = parser.add_mutually_exclusive_group()
        explicit.add_argument("--id", type=lambda x: int(x, 0), help="Single device to upgrade")
        explicit.add_argument("--list", type=ValidFile, help="File containing a list of IDs to upgrade")

    def progress_table(self):
        table = Table()
        table.add_column(f"{self._app_name}\n{self._new_ver}")
        table.add_column("Count")
        table.add_row("Updated", str(self._updated))
        table.add_row("Pending", str(len(self._pending)))
        table.add_row("Already", str(self._already))
        table.add_row("Failed", str(self._failed))
        table.add_row("No Diff", str(self._no_diff))

        if len(self._missing_diffs) > 0:
            table.add_section()
            table.add_row("Missing diffs", "\n".join(self._missing_diffs))

        meta = Table(box=None)
        meta.add_column()
        meta.add_row(table)
        meta.add_row(Status(self.state))
        meta.add_row(self.progress)

        return meta

    def state_update(self, live: Live, state: str):
        self.state = state
        live.update(self.progress_table())

    def data_progress_cb(self, offset):
        if self.task is None:
            self.state = "Writing patch file"
            self.task = self.progress.add_task("", total=len(self.patch_file))
        self.progress.update(self.task, completed=offset)

    def gateway_diff_load(self):
        assert self._single_diff is not None
        with self._single_diff.open("rb") as f:
            patch_file = f.read()

        with self._client.connection(InfuseID.GATEWAY, GatewayRequestConnectionRequest.DataType.COMMAND, 10) as _mtu:
            rpc_client = RpcClient(self._client, _mtu, InfuseID.GATEWAY)
            params = file_write_basic.request(rpc_enum_file_action.FILE_FOR_COPY, binascii.crc32(patch_file))

            print(f"Writing '{self._single_diff}' to gateway")
            hdr, _rsp = rpc_client.run_data_send_cmd(
                file_write_basic.COMMAND_ID,
                Auth.DEVICE,
                bytes(params),
                patch_file,
                None,
                file_write_basic.response.from_buffer_copy,
            )
            if hdr.return_code != 0:
                sys.exit(f"Failed to save diff file to gateway (({errno.strerror(-hdr.return_code)}))")
            print(f"'{self._single_diff}' written to gateway")

    def run_file_upload(self, live: Live, mtu: int, source: HopReceived):
        self.state_update(live, f"Uploading patch file to {source.infuse_id:016X}")
        rpc_client = RpcClient(self._client, mtu, source.infuse_id)

        params = file_write_basic.request(rpc_enum_file_action.APP_CPATCH, binascii.crc32(self.patch_file))

        hdr, _rsp = rpc_client.run_data_send_cmd(
            file_write_basic.COMMAND_ID,
            Auth.DEVICE,
            bytes(params),
            self.patch_file,
            self.data_progress_cb,
            file_write_basic.response.from_buffer_copy,
        )

        if hdr.return_code == 0:
            self._pending[source.infuse_id] = time.time() + 60

    def run_file_copy(self, live: Live, mtu: int, source: HopReceived):
        self.state_update(live, f"Copying patch file to {source.infuse_id:016X}")
        rpc_client = RpcClient(self._client, mtu, InfuseID.GATEWAY)

        params = bt_file_copy_basic.request(
            source.interface_address.val.to_rpc_struct(),
            rpc_enum_file_action.APP_CPATCH,
            0,
            len(self.patch_file),
            binascii.crc32(self.patch_file),
            1,
            3,
        )

        hdr, _rsp = rpc_client.run_standard_cmd(
            bt_file_copy_basic.COMMAND_ID,
            Auth.DEVICE,
            bytes(params),
            bt_file_copy_basic.response.from_buffer_copy,
        )
        if hdr.return_code == 0:
            self._pending[source.infuse_id] = time.time() + 60

    def run(self):
        if self._single_diff:
            self.gateway_diff_load()

        with Live(self.progress_table(), refresh_per_second=4) as live:
            for source, announce in self._client.observe_announce():
                self.state_update(live, "Scanning")
                if len(self._explicit_ids):
                    if source.infuse_id not in self._explicit_ids:
                        continue
                    if len(self._handled) == len(self._explicit_ids):
                        # We've handled all devices
                        self.state_update(live, "All devices updated")
                        return
                else:
                    if announce.application != self._app_id:
                        continue
                if source.infuse_id in self._handled:
                    continue
                v = announce.version
                v_str = f"{v.major}.{v.minor}.{v.revision}+{v.build_num:08x}"

                # Check against pending upgrades
                if source.infuse_id in self._pending:
                    if (v_str != self._new_ver) and (time.time() < self._pending[source.infuse_id]):
                        # Device could still be applying the upgrade
                        continue
                    self._pending.pop(source.infuse_id)
                    self._handled.append(source.infuse_id)
                    if v_str == self._new_ver:
                        self._updated += 1
                        result = "upgraded"
                    else:
                        self._failed += 1
                        result = "failed"
                    if self._log:
                        self._log.write(
                            f"{time.time()},0x{source.infuse_id:016x},0x{self._app_id:08x},{v_str},{result}\n"
                        )
                        self._log.flush()
                    continue

                # Already running the requested version?
                if v_str == self._new_ver:
                    self._handled.append(source.infuse_id)
                    self._already += 1
                    self.state_update(live, "Scanning")
                    if self._log:
                        self._log.write(
                            f"{time.time()},0x{source.infuse_id:016x},0x{self._app_id:08x},{v_str},already\n"
                        )
                        self._log.flush()
                    continue

                # Do we have a valid diff?
                diff_file = self._release.dir / "diffs" / f"{v_str}.bin"

                if not diff_file.exists():
                    # Is this a single diff from a different application we know about?
                    diff_file = self._release.dir / "diffs" / f"0x{announce.application:08x}" / f"{v_str}.bin"
                    if not diff_file.exists():
                        self._missing_diffs.add(v_str)
                        self._handled.append(source.infuse_id)
                        self._no_diff += 1
                        self.state_update(live, "Scanning")
                        continue

                if self._single_diff and self._single_diff != diff_file:
                    # Not the file we've copied to the gateway flash
                    self._missing_diffs.add(v_str)
                    self._handled.append(source.infuse_id)
                    self._no_diff += 1
                    self.state_update(live, "Scanning")
                    continue

                # Is signal strong enough to connect?
                if self._min_rssi and source.rssi < self._min_rssi:
                    continue

                # Load patch file
                with open(diff_file, "rb") as f:
                    self.patch_file = f.read()

                # Attempt to upload
                self.state_update(live, f"Connecting to {source.infuse_id:016X}")
                try:
                    with self._client.connection(
                        source.infuse_id, GatewayRequestConnectionRequest.DataType.COMMAND, self._conn_timeout
                    ) as mtu:
                        if self._single_diff:
                            self.run_file_copy(live, mtu, source)
                        else:
                            self.run_file_upload(live, mtu, source)

                except ConnectionRefusedError:
                    self.state_update(live, "Scanning")
                except ConnectionAbortedError:
                    self.state_update(live, "Scanning")

                if self.task is not None:
                    self.progress.remove_task(self.task)
                    self.task = None

                self.state_update(live, "Scanning")



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class last_reboot(InfuseRpcCommand, defs.last_reboot):
    @classmethod
    def add_parser(cls, parser):
        pass

    def __init__(self, args):
        pass

    def request_struct(self):
        return self.request()

    def request_json(self):
        return {}

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query reboot info ({errno.strerror(-return_code)})")
            return

        from infuse_iot.time import InfuseTime, InfuseTimeSource

        t_remote = InfuseTime.unix_time_from_epoch(response.epoch_time)

        print(f"\t     Reason: {response.reason}")
        print(f"\t   Hardware: 0x{response.hardware_flags:08x}")
        print(f"\tReboot Time: {InfuseTime.utc_time_string(t_remote)} ({InfuseTimeSource(response.epoch_time_source)})")
        print(f"\t     Uptime: {response.uptime}")
        print(f"\t    Param 1: 0x{response.param_1:08x}")
        print(f"\t    Param 2: 0x{response.param_2:08x}")
        print(f"\t     Thread: {response.thread.decode('utf-8')}")
        for idx, val in enumerate(response.esf):
            print(f"\t    ESF[{idx:2d}]: 0x{val:08x}")



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

if TYPE_CHECKING:
    from ..models.device import Device
    from ..models.downlink_message import DownlinkMessage


T = TypeVar("T", bound="RpcMessage")


@_attrs_define
class RpcMessage:
    """
    Attributes:
        created_at (datetime.datetime): The time the RPC message was created
        id (UUID): The ID of the RPC message Example: 5f4b1b2b-3b4d-4b5e-8c6f-7d8e9f0a1b2c.
        downlink_message_id (UUID): The ID of the corresponding downlink message sent to the device Example:
            7527bf1c-9868-4afd-b07d-16dc7eb7bed3.
        device (Device):
        downlink_message (DownlinkMessage):
    """

    created_at: datetime.datetime
    id: UUID
    downlink_message_id: UUID
    device: "Device"
    downlink_message: "DownlinkMessage"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        created_at = self.created_at.isoformat()

        id = str(self.id)

        downlink_message_id = str(self.downlink_message_id)

        device = self.device.to_dict()

        downlink_message = self.downlink_message.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "createdAt": created_at,
                "id": id,
                "downlinkMessageId": downlink_message_id,
                "device": device,
                "downlinkMessage": downlink_message,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.device import Device
        from ..models.downlink_message import DownlinkMessage

        d = dict(src_dict)
        created_at = isoparse(d.pop("createdAt"))

        id = UUID(d.pop("id"))

        downlink_message_id = UUID(d.pop("downlinkMessageId"))

        device = Device.from_dict(d.pop("device"))

        downlink_message = DownlinkMessage.from_dict(d.pop("downlinkMessage"))

        rpc_message = cls(
            created_at=created_at,
            id=id,
            downlink_message_id=downlink_message_id,
            device=device,
            downlink_message=downlink_message,
        )

        rpc_message.additional_properties = d
        return rpc_message

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3


from typing_extensions import Self


class Serializable:
    def to_json(self) -> dict:
        """Convert class to json dictionary"""
        raise NotImplementedError

    @classmethod
    def from_json(cls, values: dict) -> Self:
        """Reconstruct class from json dictionary"""
        raise NotImplementedError



#!/usr/bin/env python3

import keyring
import yaml

DEFAULT_NETWORK_KEY = (
    b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
    b"\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"
)


def set_api_key(api_key: str) -> None:
    """
    Save the Infuse-IoT API key to the keyring module
    """
    keyring.set_password("infuse-iot", "api-key", api_key)


def get_api_key() -> str:
    """
    Retrieve the Infuse-IoT API key from the keyring module
    """
    key = keyring.get_password("infuse-iot", "api-key")
    if key is None:
        raise FileNotFoundError("API key does not exist in keyring")
    return key


def delete_api_key() -> None:
    """
    Delete the Infuse-IoT API key from the keyring module
    """
    keyring.delete_password("infuse-iot", "api-key")


def save_network(network_id: int, network_info: str) -> None:
    """
    Save an Infuse-IoT network key to the keyring module
    """
    username = f"network-{network_id:06x}"
    keyring.set_password("infuse-iot", username, network_info)


def load_network(network_id: int) -> dict:
    """
    Retrieve an Infuse-IoT network key from the keyring module
    """
    if network_id == 0x000000:
        # Default network
        return {"id": 0, "key": DEFAULT_NETWORK_KEY}

    username = f"network-{network_id:06x}"
    key = keyring.get_password("infuse-iot", username)
    if key is None:
        raise FileNotFoundError(f"Network key {network_id:06x} does not exist in keyring")
    return yaml.safe_load(key)


def set_custom_tool_path(path: str):
    """
    Save the location of custom Infuse-IoT tools on the filesystem
    """
    keyring.set_password("infuse-iot", "custom-tools", path)


def get_custom_tool_path() -> str | None:
    """
    Retrieve the location of custom Infuse-IoT tools on the filesystem
    """
    return keyring.get_password("infuse-iot", "custom-tools")


def delete_custom_tool_path() -> None:
    """
    Delete the location of custom Infuse-IoT tools on the filesystem
    """
    return keyring.delete_password("infuse-iot", "custom-tools")


def set_custom_definitions_path(path: str):
    """
    Save the location of custom Infuse-IoT definitions on the filesystem
    """
    keyring.set_password("infuse-iot", "custom-definitions", path)


def get_custom_definitions_path() -> str | None:
    """
    Retrieve the location of custom Infuse-IoT definitions on the filesystem
    """
    return keyring.get_password("infuse-iot", "custom-definitions")


def delete_custom_definitions_path() -> None:
    """
    Delete the location of custom Infuse-IoT definitions on the filesystem
    """
    return keyring.delete_password("infuse-iot", "custom-definitions")



from enum import Enum


class KeyInterface(str, Enum):
    BT_ADV = "bt_adv"
    BT_GATT = "bt_gatt"
    SERIAL = "serial"
    UDP = "udp"

    def __str__(self) -> str:
        return str(self.value)



# Copyright (c) 2020 Teslabs Engineering S.L.
# Copyright (c) 2025 Embeint Holdings Pty Ltd
#
# SPDX-License-Identifier: Apache-2.0

import os
import platform
import re
import shutil
import subprocess
from pathlib import Path

from infuse_iot.util.soc.soc import ProvisioningInterface


class STM32Family:
    UID_ADDRESS: int
    OTP_ADDRESS: int
    SOC: str


class STM32L4X(STM32Family):
    UID_ADDRESS = 0x1FFF7590
    OTP_ADDRESS = 0x1FFF7000
    SOC = "stm32l4x"


DEVICE_ID_MAPPING = {
    0x435: STM32L4X,
    0x462: STM32L4X,
    0x464: STM32L4X,
}


class Interface(ProvisioningInterface):
    @staticmethod
    def programmer_path() -> Path:
        """
        Obtain path of the STM32CubeProgrammer CLI tool.
        From Zephyr runner file: stm32cubeprogrammer.py
        """

        if platform.system() == "Linux":
            cmd = shutil.which("STM32_Programmer_CLI")
            if cmd is not None:
                return Path(cmd)

            return (
                Path.home()
                / "STMicroelectronics"
                / "STM32Cube"
                / "STM32CubeProgrammer"
                / "bin"
                / "STM32_Programmer_CLI"
            )

        if platform.system() == "Windows":
            cli = Path("STMicroelectronics") / "STM32Cube" / "STM32CubeProgrammer" / "bin" / "STM32_Programmer_CLI.exe"
            x86_path = Path(os.environ["PROGRAMFILES(X86)"]) / cli
            if x86_path.exists():
                return x86_path

            return Path(os.environ["PROGRAMW6432"]) / cli

        if platform.system() == "Darwin":
            return (
                Path("/Applications")
                / "STMicroelectronics"
                / "STM32Cube"
                / "STM32CubeProgrammer"
                / "STM32CubeProgrammer.app"
                / "Contents"
                / "MacOs"
                / "bin"
                / "STM32_Programmer_CLI"
            )

        raise NotImplementedError("Could not determine STM32_Programmer_CLI path")

    @staticmethod
    def _get_device_id() -> int:
        """Retrieve STM32 device ID (SoC series identifier)"""
        stm_path = Interface.programmer_path()

        cmd_base = [str(stm_path), "--connect", "port=SWD"]

        # Get the device identifier (SoC Series)
        res = subprocess.run(cmd_base, capture_output=True, check=True)
        output = res.stdout.decode("utf-8")

        match = re.search(r"Device ID *: (0x[0-9a-fA-F]*)", output, re.MULTILINE)
        if match is None:
            raise RuntimeError("No match for 'Device ID' found in output")
        return int(match.group(1), 16)

    def __init__(self):
        # Get the device identifier (SoC Series)
        self._device_id = self._get_device_id()
        self._family = DEVICE_ID_MAPPING[self._device_id]
        self._cli = Interface.programmer_path()

    @property
    def soc_name(self) -> str:
        return self._family.SOC

    @property
    def unique_device_id_len(self) -> int:
        return 12

    def unique_device_id(self) -> int:
        """Retrieve 96 bit STM32 unique device ID"""
        cmd_base = [str(self._cli), "--connect", "port=SWD"]

        # Read the 96 bit (12 byte) UID
        cmd_read = cmd_base + ["-r32", hex(self._family.UID_ADDRESS), "12"]
        res = subprocess.run(cmd_read, capture_output=True, check=True)
        output = res.stdout.decode("utf-8")

        match = re.search(r"([0-9a-fA-F]{8}) ([0-9a-fA-F]{8}) ([0-9a-fA-F]{8})", output, re.MULTILINE)
        if match is None:
            raise RuntimeError("Unique device ID not found in output")

        word0 = int(match.group(1), 16)
        word1 = int(match.group(2), 16)
        word2 = int(match.group(3), 16)

        # Byte order matches that returned by `hwinfo_stm32`
        id_bytes = word2.to_bytes(4, "big") + word1.to_bytes(4, "big") + word0.to_bytes(4, "big")
        return int.from_bytes(id_bytes, "big")

    @staticmethod
    def unique_device_id_fields(uid: int) -> dict:
        uid_bytes = uid.to_bytes(12, "big")
        fields = {
            "wafer_x": int.from_bytes(uid_bytes[10:12], "big"),
            "wafer_y": int.from_bytes(uid_bytes[8:10], "big"),
            "wafer_number": uid_bytes[7],
            "lot_number": uid_bytes[0:7].decode("utf-8"),
        }
        return fields

    def read_provisioned_data(self, num: int) -> bytes:
        cmd_base = [str(self._cli), "--connect", "port=SWD"]

        cmd_read = cmd_base + ["-r32", hex(self._family.OTP_ADDRESS), str(num)]
        res = subprocess.run(cmd_read, capture_output=True, check=True)
        output = res.stdout.decode("utf-8")

        # Searching for patterns like the following:
        #
        # 0x1FFF7000 : FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF
        # 0x1FFF7010 : FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF
        # 0x1FFF7020 : FFFFFFFF FFFFFFFF
        byte_re = re.compile(
            r"0x[0-9a-fA-F]{8} : ([0-9a-fA-F]{8}) ?([0-9a-fA-F]{8})? ?([0-9a-fA-F]{8})? ?([0-9a-fA-F]{8})?",
            re.MULTILINE,
        )
        byte_string_lines = byte_re.findall(output)

        # Parse into a byte string
        byte_output = b""
        for line in byte_string_lines:
            for chunk in line:
                if chunk == "":
                    continue
                # Byte order needs to be reversed
                big_endian = bytes.fromhex(chunk)
                byte_output += bytes(reversed(big_endian))

        # Return the requested number of bytes
        return byte_output[:num]

    def write_provisioning_data(self, data: bytes):
        cmd_base = [str(self._cli), "--connect", "port=SWD"]

        # Pad to 32bit alignment
        if mod := len(data) % 4:
            data += b"\xff" * (4 - mod)

        # Convert bytes to 4 byte words
        words = [int.from_bytes(data[i : i + 4], "little") for i in range(0, len(data), 4)]

        cmd_write = cmd_base + ["-w32", hex(self._family.OTP_ADDRESS)] + [f"0x{w:08x}" for w in words]
        subprocess.run(cmd_write, capture_output=True, check=True)

    def close(self):
        cmd_reset = [str(self._cli), "--connect", "port=SWD", "-rst"]
        subprocess.run(cmd_reset, capture_output=True, check=True)



import os
import subprocess
import sys

import infuse_iot.version

assert "TOXTEMPDIR" in os.environ, "you must run these tests using tox"


def test_main():
    # A quick check that the package can be executed as a module which
    # takes arguments, using e.g. "python3 -m west --version" to
    # produce the same results as "west --version", and that both are
    # sane (i.e. the actual version number is printed instead of
    # simply an error message to stderr).

    output_as_module = subprocess.check_output([sys.executable, "-m", "infuse_iot", "--version"]).decode()
    output_directly = subprocess.check_output(["infuse", "--version"]).decode()
    assert infuse_iot.version.__version__ in output_as_module
    assert output_as_module == output_directly



#!/usr/bin/env python3

import ctypes

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.util.ctypes import UINT16_MAX
from infuse_iot.zephyr.errno import errno


class infuse_states_update(InfuseRpcCommand, defs.infuse_states_update):
    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--state", "-s", type=int, required=True, help="State ID to update")
        option = parser.add_mutually_exclusive_group(required=True)
        option.add_argument("--set", action="store_true", help="Enable the state permanently")
        option.add_argument("--clear", action="store_true", help="Disable the state")
        option.add_argument("--timeout", type=int, help="Enable the state for a duration (seconds)")

    def __init__(self, args):
        self.args = args

    def request_struct(self):
        class request(ctypes.LittleEndianStructure):
            _fields_ = [
                ("num", ctypes.c_uint8),
                ("state", defs.rpc_struct_infuse_state),
            ]

        timeout = 0 if self.args.set else (UINT16_MAX if self.args.clear else self.args.timeout)
        state = defs.rpc_struct_infuse_state(self.args.state, timeout)

        return request(1, state)

    def request_json(self):
        return {}

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to update state {self.args.state} ({errno.strerror(-return_code)})")
        else:
            print(f"Updated state {self.args.state}")



#!/usr/bin/env python3

import ctypes
import random

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.epacket.packet import Auth
from infuse_iot.util.argparse import ValidFile
from infuse_iot.util.ctypes import bytes_to_uint8
from infuse_iot.zephyr.errno import errno


class challenge_response_header(ctypes.LittleEndianStructure):
    _fields_ = [
        ("cloud_public_key", 32 * ctypes.c_uint8),
        ("device_public_key", 32 * ctypes.c_uint8),
        ("network_id", ctypes.c_uint32),
        ("challenge_response_type", ctypes.c_uint8),
    ]
    _pack_ = 1


class challenge_response_basic_encr(ctypes.LittleEndianStructure):
    _fields_ = [
        ("nonce", 12 * ctypes.c_uint8),
        ("challenge", 16 * ctypes.c_uint8),
        ("identity", 16 * ctypes.c_uint8),
        ("device_id", ctypes.c_uint64),
        ("tag", 16 * ctypes.c_uint8),
    ]
    _pack_ = 1


class challenge_response_basic(ctypes.LittleEndianStructure):
    _fields_ = [
        ("challenge", 16 * ctypes.c_uint8),
        ("identity", 16 * ctypes.c_uint8),
        ("device_id", ctypes.c_uint64),
    ]
    _pack_ = 1


class security_state(InfuseRpcCommand, defs.security_state):
    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--pem", type=ValidFile, help="Cloud .pem file for identity validation")

    def __init__(self, args):
        self.challenge = bytes_to_uint8(random.randbytes(16))
        self.pem = args.pem

    def auth_level(self):
        return Auth.NETWORK

    def request_struct(self):
        return self.request(self.challenge)

    def _decrypt_response(self, response):
        rsp = response.response

        rb = bytes(response.response)
        with self.pem.open("r") as f:
            cloud_private_key = serialization.load_pem_private_key(f.read().encode("utf-8"), password=None)
            assert isinstance(cloud_private_key, x25519.X25519PrivateKey)
        device_public_key = x25519.X25519PublicKey.from_public_bytes(bytes(response.device_public_key))
        shared_secret = cloud_private_key.exchange(device_public_key)
        hkdf = HKDF(
            algorithm=hashes.SHA256(),
            length=32,
            salt=0x1234.to_bytes(4, "little"),
            info=b"sign",
        )
        sign_key = hkdf.derive(shared_secret)

        cipher = ChaCha20Poly1305(sign_key)
        plaintext = cipher.decrypt(bytes(rsp.nonce), rb[12:], bytes(response.header))
        return challenge_response_basic.from_buffer_copy(plaintext)

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query current time ({errno.strerror(-return_code)})")
            return

        # Decrypt identity information
        print("Security State:")
        print(f"\tDevice Public Key: {bytes(response.device_public_key).hex()}")
        print(f"\t Cloud Public Key: {bytes(response.cloud_public_key).hex()}")
        print(f"\t          Network: 0x{response.network_id:06x}")
        if self.pem is None:
            print("\t         Identity: Cannot validate")
        else:
            challenge_rsp = self._decrypt_response(response)

            valid_response = bytes(challenge_rsp.challenge) == self.challenge

            print(f"\t         Response: {'Valid' if valid_response else 'Invalid'}")
            print(f"\t        Device ID: {challenge_rsp.device_id:016x}")
            print(f"\t  Identity Secret: {bytes(challenge_rsp.identity).hex()}")
